r"""Auxiliary functions for writing to `TensorBoard <https://www.tensorflow.org/tensorboard>`_ event files."""

import re
from typing import Callable, Optional, Union

from torch import Tensor
from torch.utils.tensorboard import SummaryWriter

from ..core.image import normalize_image
from ..core.typing import TensorCollection


RE_CHANNEL_INDEX = re.compile(r"\{c(:[^}]+)?\}")


def escape_channel_index_format_string(tag: str) -> str:
    r"""Escape image channel index format before str.format() call."""
    return RE_CHANNEL_INDEX.sub(r"{{c\1}}", tag)


def add_summary_image(
    writer: SummaryWriter,
    tag: str,
    image: Tensor,
    global_step: Optional[int] = None,
    walltime: Optional[float] = None,
    image_transform: Union[bool, Callable[[str, Tensor], Tensor]] = True,
    rescale_transform: Union[bool, Callable[[str, Tensor], Tensor]] = False,
    channel_offset: int = 0,
) -> None:
    r"""Add image to TensorBoard summary.

    Args:
        writer: TensorBoard writer with open event file.
        tag: Image tag passed to ``writer.add_image``.
        image: Image data tensor.
        global_step: Global step value to record.
        walltime: Optional override default walltime seconds.
        image_transform: Callable used to extract a 2D image tensor of shape
            ``(C, Y, X)`` from ``data``. When a multi-channel tensor is returnd,
            each channel is saved as separate image to the TensorBoard event file.
            with channel index appended to the ``tag`` separated with an underscore ``_``.
            By default, the central slice of the first image in the batch is extracted.
            The first argument is the ``tag`` of the image tensor. This can be used to
            differently process different images by the same callable transform.
        rescale_transform: Image intensities must be in the closed interval ``[0, 1]``.
            When ``False``, this must already the case, e.g., as part of ``image_transform``.
            When ``True``, images with values outside this interval are rescaled unless
            the ``tag`` matches "y" or "y_pred".
        channel_offset: Offset to add to channel index in tag format string.

    """
    img = image.detach()
    if image_transform is True:
        image_transform = first_central_image_slice
    if image_transform not in (False, None):
        img = image_transform(tag, img)
    if rescale_transform is True:
        rescale_transform = normalize_summary_image
    if rescale_transform not in (False, None):
        img = rescale_transform(tag, img)
    if img.ndim != 3:
        raise AssertionError("add_summary_image() transformed tensor must have shape (C, H, W)")
    if img.shape[0] > 1 and RE_CHANNEL_INDEX.search(tag) is None:
        tag = tag + "/{c}"
    kwargs = dict(global_step=global_step, walltime=walltime)
    for c in range(img.shape[0]):
        writer.add_image(tag.format(c=c + channel_offset), img.narrow(0, c, 1), **kwargs)


def add_summary_images(
    writer: SummaryWriter,
    prefix: str,
    images: TensorCollection,
    global_step: Optional[int] = None,
    walltime: Optional[float] = None,
    image_transform: Union[bool, Callable[[str, Tensor], Tensor]] = True,
    rescale_transform: Union[bool, Callable[[str, Tensor], Tensor]] = False,
    channel_offset: int = 0,
) -> None:
    r"""Add slices of image tensors to TensorBoard summary.

    Args:
        writer: TensorBoard writer with open event file.
        prefix: Prefix string for TensorBoard tags.
        images: Possibly nested dictionary and/or sequence of image tensors.
        global_step: Global step value to record.
        walltime: Optional override default walltime seconds.
        image_transform: Callable used to extract a 2D image tensor of shape
            ``(C, Y, X)`` from each image. When a multi-channel tensor is returnd (C > 1),
            each channel is saved as separate image to the TensorBoard event file.
            By default, the central slice of the first image in the batch is extracted.
            The first argument is the name of the image tensor if applicable, or an empty
            string otherwise. This can be used to differently process different images.
        rescale_transform: Image intensities must be in the closed interval ``[0, 1]``.
            Set to ``False``, if this is already the case or when a custom
            ``image_transform`` is used which normalizes the image intensities.
        channel_offset: Offset to add to channel index in tag format string.

    """
    if prefix is None:
        prefix = ""
    if not isinstance(images, dict):
        images = {str(i): value for i, value in enumerate(images)}
    for name, value in images.items():
        if isinstance(value, Tensor):
            add_summary_image(
                writer,
                prefix + name,
                value,
                global_step=global_step,
                walltime=walltime,
                image_transform=image_transform,
                rescale_transform=rescale_transform,
                channel_offset=channel_offset,
            )
        else:
            add_summary_images(
                writer,
                prefix + name + "/",
                value,
                global_step=global_step,
                walltime=walltime,
                image_transform=image_transform,
                rescale_transform=rescale_transform,
                channel_offset=channel_offset,
            )


def central_image_slices(tag: str, data: Tensor, start: int = 0, length: int = -1) -> Tensor:
    if data.ndim < 4 or data.shape[1] != 1:
        raise AssertionError(
            "central_image_slices() expects image tensors of shape (N, 1, ..., Y, X)"
        )
    data = data.narrow(1, 0, 1).squeeze(1)
    for dim in range(1, data.ndim - 2):
        data = data.narrow(dim, data.shape[dim] // 2, 1)
    for dim in range(1, data.ndim - 2):
        data = data.squeeze(dim)
    if length < 1:
        length = data.shape[0]
    length = min(length, data.shape[0])
    return data.narrow(0, start, length)


def all_central_image_slices(tag: str, data: Tensor) -> Tensor:
    r"""Extract central slice from each scalar image in batch."""
    return central_image_slices(tag, data, start=0, length=-1)


def first_central_image_slice(tag: str, data: Tensor) -> Tensor:
    r"""Extract central slice of first image in batch."""
    return central_image_slices(tag, data, start=0, length=1)


def normalize_summary_image(tag: str, data: Tensor) -> Tensor:
    r"""Linearly rescale image values to [0, 1]."""
    return normalize_image(data, mode="unit")



r"""Free-form deformation (FFD) regularization terms."""

from torch import Tensor

from .base import BSplineLoss
from . import functional as L


class BSplineBending(BSplineLoss):
    r"""Bending energy of cubic B-spline free form deformation."""

    def forward(self, params: Tensor) -> Tensor:
        r"""Evaluate loss term for given free form deformation parameters."""
        return L.bending_loss(params, mode="bspline", stride=self.stride, reduction=self.reduction)


BSplineBendingEnergy = BSplineBending



r"""Common auxiliary functions for working with VTK data structures."""

from typing import Generator

from vtk import vtkIdList, vtkPolyData


def iter_id_list(ids: vtkIdList, start: int = 0, stop: int = None) -> Generator[int, None, None]:
    r"""Iterate over IDs in vtkIdList."""
    if stop is None:
        stop = ids.GetNumberOfIds()
    elif stop < 0:
        stop = stop % (ids.GetNumberOfIds() + 1)
    for idx in range(start, stop):
        yield ids.GetId(idx)


def iter_cell_point_ids(
    polydata: vtkPolyData, cell_id: int, start: int = 0, stop: int = None
) -> Generator[int, None, None]:
    r"""Iterate over IDs of points that a given vtkPolyData cell is made up of."""
    point_ids = vtkIdList()
    polydata.GetCellPoints(cell_id, point_ids)
    yield from iter_id_list(point_ids, start, stop)


def iter_point_cell_ids(
    polydata: vtkPolyData, point_id: int, start: int = 0, stop: int = None
) -> Generator[int, None, None]:
    r"""Iterate over IDs of vtkPolyData cells that contain a specified point."""
    cell_ids = vtkIdList()
    polydata.GetPointCells(point_id, cell_ids)
    yield from iter_id_list(cell_ids, start, stop)



r"""Pooling layers."""

from functools import partial
from typing import Any, Callable, Mapping, Optional, Sequence, Type, Union

from torch import Tensor, nn
from torch.nn import Module

from .lambd import LambdaLayer


PoolFunc = Callable[[Tensor], Tensor]
PoolArg = Union[PoolFunc, str, Mapping, Sequence, None]


POOLING_TYPES = {
    "avg": (nn.AvgPool1d, nn.AvgPool2d, nn.AvgPool3d),
    "avgpool": (nn.AvgPool1d, nn.AvgPool2d, nn.AvgPool3d),
    "adaptiveavg": (nn.AdaptiveAvgPool1d, nn.AdaptiveAvgPool2d, nn.AdaptiveAvgPool3d),
    "adaptiveavgpool": (
        nn.AdaptiveAvgPool1d,
        nn.AdaptiveAvgPool2d,
        nn.AdaptiveAvgPool3d,
    ),
    "adaptivemax": (nn.AdaptiveMaxPool1d, nn.AdaptiveMaxPool2d, nn.AdaptiveMaxPool3d),
    "adaptivemaxpool": (
        nn.AdaptiveMaxPool1d,
        nn.AdaptiveMaxPool2d,
        nn.AdaptiveMaxPool3d,
    ),
    "max": (nn.MaxPool1d, nn.MaxPool2d, nn.MaxPool3d),
    "maxpool": (nn.MaxPool1d, nn.MaxPool2d, nn.MaxPool3d),
    "maxunpool": (nn.MaxUnpool1d, nn.MaxUnpool2d, nn.MaxUnpool3d),
    "identity": nn.Identity,
}


def pooling(
    arg: PoolArg,
    *args: Any,
    spatial_dims: Optional[int] = None,
    **kwargs,
) -> Module:
    r"""Get pooling layer.

    Args:
        arg: Custom pooling function or module, or name of pooling layer with optional keyword arguments.
            When ``arg`` is a callable but not of type ``torch.nn.Module``, it is wrapped in a ``PoolLayer``.
            If ``None`` or 'identity', an instance of ``torch.nn.Identity`` is returned.
        spatial_dims: Number of spatial dimensions of input tensors.
        args: Arguments to pass to init function of pooling layer. If ``arg`` is a callable, the given arguments
            are passed to the function each time it is called as arguments.
        kwargs: Additional keyword arguments for initialization of pooling layer. Overrides keyword arguments given as
            second tuple item when ``arg`` is a ``(name, kwargs)`` tuple instead of a string. When ``arg`` is a callable,
            the keyword arguments are passed each time the pooling function is called.

    Returns:
        Pooling layer instance.

    """
    if isinstance(arg, Module) and not args and not kwargs:
        return arg
    if callable(arg):
        return PoolLayer(arg, *args, **kwargs)
    pool_name = "identity"
    pool_args = {}
    if isinstance(arg, str):
        pool_name = arg
    elif isinstance(arg, Mapping):
        pool_name = arg.get("name")
        if not pool_name:
            raise ValueError("pooling() 'arg' map must contain 'name'")
        if not isinstance(pool_name, str):
            raise TypeError("pooling() 'name' must be str")
        pool_args = {key: value for key, value in arg.items() if key != "name"}
    elif isinstance(arg, Sequence):
        if len(arg) != 2:
            raise ValueError("pooling() 'arg' sequence must have length two")
        pool_name, pool_args = arg
        if not isinstance(pool_name, str):
            raise TypeError("pooling() first 'arg' sequence argument must be str")
        if not isinstance(pool_args, dict):
            raise TypeError("pooling() second 'arg' sequence argument must be dict")
        pool_args = pool_args.copy()
    elif arg is not None:
        raise TypeError("pooling() 'arg' must be str, mapping, 2-tuple, callable, or None")
    pool_type: Union[Type[Module], Sequence[Type[Module]]] = POOLING_TYPES.get(pool_name.lower())
    if pool_type is None:
        raise ValueError(f"pooling() unknown pooling layer {pool_name!r}")
    if isinstance(pool_type, Sequence):
        if spatial_dims is None:
            raise ValueError(f"pooling() 'spatial_dims' required for pooling layer {pool_name!r}")
        try:
            pool_type = pool_type[spatial_dims - 1]
        except IndexError:
            pool_type = None
        if pool_type is None:
            raise ValueError(f"pooling({pool_name!r}) does not support spatial_dims={spatial_dims}")
    pool_args.update(kwargs)
    module = pool_type(*args, **pool_args)
    return module


def pool_layer(*args, **kwargs) -> Module:
    return pooling(*args, **kwargs)


class PoolLayer(LambdaLayer):
    r"""Pooling layer."""

    def __init__(
        self,
        arg: PoolArg,
        *args: Any,
        spatial_dims: Optional[int] = None,
        **kwargs,
    ) -> None:
        if callable(arg):
            pool = partial(arg, *args, **kwargs) if args or kwargs else arg
        else:
            pool = pooling(arg, *args, spatial_dims=spatial_dims, **kwargs)
        super().__init__(pool)



r"""Generic spatial transformation model."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Dict, List, Mapping, Optional, Union

import torch
from torch import Tensor
from torch.nn import ModuleDict

from deepali.core.affine import euler_rotation_angles
from deepali.core.affine import euler_rotation_matrix
from deepali.core.config import DataclassConfig
from deepali.core.grid import Grid
from deepali.core.linalg import quaternion_to_rotation_matrix
from deepali.core.linalg import rotation_matrix_to_quaternion
from deepali.core.typing import ScalarOrTuple

from .bspline import FreeFormDeformation, StationaryVelocityFreeFormDeformation
from .composite import SequentialTransform
from .linear import AnisotropicScaling, EulerRotation, QuaternionRotation
from .linear import HomogeneousTransform, Shearing, Translation
from .nonrigid import DisplacementFieldTransform, StationaryVelocityFieldTransform


ParamsDict = Mapping[str, Tensor]


AFFINE_NAMES = {
    "A": "affine",
    "K": "shearing",
    "T": "translation",
    "R": "rotation",
    "S": "scaling",
    "Q": "quaternion",
}
r"""Names of elementary affine transformation child modules.

The dictionary key is the letter used in :attr:`TransformConfig.affine_model`, i.e.,

- ``A``: ``"affine"``
- ``K``: ``"shearing"``
- ``T``: ``"translation"``
- ``R``: ``"rotation"``
- ``S``: ``"scaling"``
- ``Q``: ``"quaternion"``

"""

AFFINE_TRANSFORMS = {
    "A": HomogeneousTransform,
    "K": Shearing,
    "T": Translation,
    "R": EulerRotation,
    "S": AnisotropicScaling,
    "Q": QuaternionRotation,
}
r"""Types of elementary affine transformations.

The dictionary key is the letter used in :attr:`TransformConfig.affine_model`, i.e.,

- ``A``: :class:`.HomogeneousTransform`
- ``K``: :class:`.Shearing`
- ``T``: :class:`.Translation`
- ``R``: :class:`.EulerRotation`
- ``S``: :class:`.AnisotropicScaling`
- ``Q``: :class:`.QuaternionRotation`

"""

NONRIGID_TRANSFORMS = {
    "DDF": DisplacementFieldTransform,
    "FFD": FreeFormDeformation,
    "SVF": StationaryVelocityFieldTransform,
    "SVFFD": StationaryVelocityFreeFormDeformation,
}
r"""Types of non-rigid transformations.

The dictionary key is the string used in :attr:`TransformConfig.transform`, i.e.,

- ``DDF``: :class:`.DisplacementFieldTransform`
- ``FFD``: :class:`.FreeFormDeformation`
- ``SVF``: :class:`.StationaryVelocityFieldTransform`
- ``SVFFD``: :class:`.StationaryVelocityFreeFormDeformation`

"""

VALID_COMPONENTS = ("Affine",) + tuple(NONRIGID_TRANSFORMS.keys())
r"""Valid transformation names in :attr:`TransformConfig.transform` string value.

This includes "Affine" and all keys of the :data:`NONRIGID_TRANSFORMS` dictionary.

"""


def transform_components(model: str) -> List[str]:
    r"""Non-rigid component of transformation or ``None`` if it is a linear transformation."""
    return model.split(" o ")


def valid_transform_model(
    model: str, max_affine: Optional[int] = None, max_nonrigid: Optional[int] = None
) -> bool:
    r"""Whether given string denotes a valid transformation model."""
    components = transform_components(model)
    num_affine = 0
    num_nonrigid = 0
    for component in components:
        if component not in VALID_COMPONENTS:
            return False
        if component == "Affine":
            num_affine += 1
        else:
            num_nonrigid += 1
    if len(components) < 1:
        return False
    if max_affine is not None and num_affine > max_affine:
        return False
    if max_nonrigid is not None and num_nonrigid > max_nonrigid:
        return False
    return True


def has_affine_component(model: str) -> bool:
    r"""Whether transformation model includes an affine component."""
    return "Affine" in transform_components(model)


def has_nonrigid_component(model: str) -> bool:
    r"""Whether transformation model includes a non-rigid component."""
    return nonrigid_components(model)


def nonrigid_components(model: str) -> List[str]:
    r"""Non-rigid components of transformation model."""
    return [comp for comp in transform_components(model) if comp in NONRIGID_TRANSFORMS]


def affine_first(model: str) -> bool:
    r"""Whether transformation applies affine component first."""
    components = transform_components(model)
    assert components, "must contain at least one transformation component"
    return components[-1] == "Affine"


@dataclass
class TransformConfig(DataclassConfig):
    r"""Configuration of generic spatial transformation model."""

    transform: str = "Affine o SVF"
    r"""String encoding of spatial transformation model to use.

    The linear transforms making up the ``Affine`` component are defined by :attr:`affine_model`.

    The non-rigid component can be one of the following:

    - ``DDF``: :class:`.DisplacementFieldTransform`
    - ``FFD``: :class:`.FreeFormDeformation`
    - ``SVF``: :class:`.StationaryVelocityFieldTransform`
    - ``SVFFD``: :class:`.StationaryVelocityFreeFormDeformation`

    """

    affine_model: str = "TRS"  # same as "T o R o S"
    r"""String encoding of composition of elementary linear transformations.

    The string value of this configuration entry can be in one of two forms:

    - Matrix notation: Each letter is a factor in the sequence of matrix-matrix products.
    - Function composition: Use deliminator " o " between transformations to denote composition.

    Valid elementary linear transform identifiers are:

    - ``A``: :class:`.HomogeneousTransform`
    - ``K``: :class:`.Shearing`
    - ``T``: :class:`.Translation`
    - ``R``: :class:`.EulerRotation`
    - ``S``: :class:`.AnisotropicScaling`
    - ``Q``: :class:`.QuaternionRotation`

    """

    rotation_model: str = "ZXZ"
    r"""Order of elementary Euler rotations.

    This configuration value is only used when :attr:`affine_model` contains an :class:`EulerRotation`
    denoted by letter "R". Valid values are "ZXZ", "XZX", ... (cf. :func:`.core.affine.euler_rotation_matrix`).

    """

    control_point_spacing: ScalarOrTuple[int] = 1
    r"""Control point spacing of non-rigid transformations.

    The spacing must be given in voxel units of the grid domain with respect to
    which the transformations are defined.

    """

    scaling_and_squaring_steps: int = 6
    r"""Number of scaling and squaring steps in case of a stationary velocity field transform."""

    flip_grid_coords: bool = False
    r"""Whether predicted transformation parameters are with respect to a grid
        with point coordinates in the order (..., x) instead of (x, ...)."""

    def _finalize(self, parent: Path) -> None:
        r"""Finalize parameters after loading these from input file."""
        super()._finalize(parent)


class GenericSpatialTransform(SequentialTransform):
    r"""Configurable generic spatial transformation."""

    def __init__(
        self,
        grid: Grid,
        params: Optional[Union[bool, Callable[..., ParamsDict], ParamsDict]] = True,
        config: Optional[TransformConfig] = None,
    ) -> None:
        r"""Initialize spatial transformation."""
        if (
            params not in (None, False, True)
            and not callable(params)
            and not isinstance(params, Mapping)
        ):
            raise TypeError(
                f"{type(self).__name__}() 'params' must be bool, callable, dict, or None"
            )
        if config is None:
            config = getattr(params, "config", None)
            if config is None:
                raise AssertionError(
                    f"{type(self).__name__}() 'config' or 'params.config' required"
                )
            if not isinstance(config, TransformConfig):
                raise TypeError(f"{type(self).__name__}() 'params.config' must be TransformConfig")
        elif not isinstance(config, TransformConfig):
            raise TypeError(f"{type(self).__name__}() 'config' must be TransformConfig")
        if not valid_transform_model(config.transform, max_affine=1, max_nonrigid=1):
            raise ValueError(
                f"{type(self).__name__}() 'config.transform' invalid or not supported: {config.transform}"
            )
        modules = ModuleDict()
        # Initialize affine components
        if has_affine_component(config.transform):
            for key in reversed(config.affine_model.replace(" o ", "")):
                key = key.upper()
                if key not in AFFINE_TRANSFORMS:
                    raise ValueError(
                        f"{type(self).__name__}() invalid character '{key}' in 'config.affine_model'"
                    )
                name = AFFINE_NAMES[key]
                if name in modules:
                    raise NotImplementedError(
                        f"{type(self).__name__}() 'config.affine_model' must contain each elementary"
                        f" transform at most once, but encountered key '{key}' more than once."
                    )
                kwargs = dict(grid=grid, params=params if isinstance(params, bool) else None)
                if key == "R":
                    kwargs["order"] = config.rotation_model
                modules[name] = AFFINE_TRANSFORMS[key](**kwargs)
        # Initialize non-rigid component
        nonrigid_models = nonrigid_components(config.transform)
        if len(nonrigid_models) > 1:
            raise ValueError(
                f"{type(self).__name__}() 'config.transform' must contain at most one non-rigid component"
            )
        if nonrigid_models:
            nonrigid_model = nonrigid_models[0]
            nonrigid_params = params if isinstance(params, bool) else None
            nonrigid_kwargs = dict(grid=grid, params=nonrigid_params)
            NonRigidTransform = NONRIGID_TRANSFORMS[nonrigid_model]
            if nonrigid_model in ("DDF", "SVF") and config.control_point_spacing > 1:
                size = grid.size_tensor()
                stride = torch.tensor(config.control_point_spacing).to(size)
                size = size.div(stride).ceil().long()
                nonrigid_kwargs["grid"] = grid.resize(size)
            if nonrigid_model == "SVF":
                nonrigid_kwargs["steps"] = config.scaling_and_squaring_steps
            if nonrigid_model in ("FFD", "SVFFD"):
                nonrigid_kwargs["stride"] = config.control_point_spacing
            _modules = ModuleDict({"nonrigid": NonRigidTransform(**nonrigid_kwargs)})
            if affine_first(config.transform):
                modules.update(_modules)
            else:
                _modules.update(modules)
                modules = _modules
        # Set parameters of transformation if given as dictionary
        if isinstance(params, Mapping):
            for name, transform in self.named_transforms():
                transform.data_(params[name])
        # Insert transformations in order of composition
        super().__init__(grid, modules)
        self.config = config
        self.params = params if callable(params) else None

    def _data(self) -> Dict[str, Tensor]:
        r"""Get most recent transformation parameters."""
        if not self._transforms:
            return {}
        params = self.params
        if params is None:
            params = {}
            for name, transform in self.named_transforms():
                params[name] = transform.data()
            return params
        if isinstance(params, GenericSpatialTransform):
            return {}  # transforms are individually linked to their counterparts
        if callable(params):
            args, kwargs = self.condition()
            pred = params(*args, **kwargs)
            if not isinstance(pred, Mapping):
                raise TypeError(
                    f"{type(self).__name__} 'params' callable return value must be a Mapping"
                )
        elif isinstance(params, Mapping):
            pred = params
        else:
            raise TypeError(
                f"{type(self).__name__} 'params' attribute must be a"
                " callable, Mapping, linked GenericSpatialTransform, or None"
            )
        data = {}
        flip_grid_coords = self.config.flip_grid_coords
        if "affine" in self._transforms:
            matrix = pred["affine"]
            assert isinstance(matrix, Tensor)
            assert matrix.ndim >= 2
            D = matrix.shape[-2]
            assert matrix.shape[-1] == D + 1
            if flip_grid_coords:
                matrix[..., :D, :D] = matrix[..., :D, :D].flip((1, 2))
                matrix[..., :D, -1] = matrix[..., :D, -1].flip(-1)
            data["affine"] = matrix
        if "translation" in self._transforms:
            if "translation" in pred:
                offset = pred["translation"]
            else:
                offset = pred["offset"]
            assert isinstance(offset, Tensor)
            if flip_grid_coords:
                offset = offset.flip(-1)
            data["translation"] = offset
        if "rotation" in self._transforms:
            if "rotation" in pred:
                angles = pred["rotation"]
            else:
                angles = pred["angles"]
            assert isinstance(angles, Tensor)
            if flip_grid_coords:
                rotmodel = self.config.rotation_model
                rotation = euler_rotation_matrix(angles, order=rotmodel).flip((1, 2))
                angles = euler_rotation_angles(rotation, order=rotmodel)
            data["rotation"] = angles
        if "scaling" in self._transforms:
            if "scaling" in pred:
                scales = pred["scaling"]
            else:
                scales = pred["scales"]
            assert isinstance(scales, Tensor)
            if flip_grid_coords:
                scales = scales.flip(-1)
            data["scaling"] = scales
        if "quaternion" in self._transforms:
            q = pred["quaternion"]
            assert isinstance(q, Tensor)
            if flip_grid_coords:
                m = quaternion_to_rotation_matrix(q)
                m = m.flip((1, 2))
                q = rotation_matrix_to_quaternion(m)
            data["quaternion"] = q
        if "nonrigid" in self._transforms:
            if "nonrigid" in pred:
                vfield = pred["nonrigid"]
            else:
                vfield = pred["vfield"]
            assert isinstance(vfield, Tensor)
            if flip_grid_coords:
                vfield = vfield.flip(1)
            data["nonrigid"] = vfield
        return data

    def inverse(self, link: bool = False, update_buffers: bool = False) -> GenericSpatialTransform:
        r"""Get inverse of this transformation.

        Args:
            link: Whether the inverse transformation keeps a reference to this transformation.
                If ``True``, the ``update()`` function of the inverse function will not recompute
                shared parameters (e.g., parameters obtained by a callable neural network), but
                directly access the parameters from this transformation. Note that when ``False``,
                the inverse transformation will still share parameters, modules, and buffers with
                this transformation, but these shared tensors may be replaced by a call of ``update()``
                (which is implicitly called as pre-forward hook when ``__call__()`` is invoked).
            update_buffers: Whether buffers of inverse transformation should be updated after creating
                the shallow copy. If ``False``, the ``update()`` function of the returned inverse
                transformation has to be called before it is used.

        Returns:
            Shallow copy of this transformation which computes and applied the inverse transformation.
            The inverse transformation will share the parameters with this transformation. Not all
            transformations may implement this functionality.

        Raises:
            NotImplementedError: When a transformation does not support sharing parameters with its inverse.

        """
        inv = super().inverse(link=link, update_buffers=update_buffers)
        if link:
            inv.params = self
        return inv

    def update(self) -> GenericSpatialTransform:
        r"""Update transformation parameters."""
        if self.params is not None:
            params = self._data()
            for k, p in params.items():
                transform = self._transforms[k]
                transform.data_(p)
        super().update()
        return self



from collections import namedtuple
from typing import Any, Iterable, Mapping, NamedTuple, Optional, Sequence, Tuple, Union, overload

import torch
from torch import Size, Tensor

from .typing import ScalarOrTuple


def get_namedtuple_item(self: NamedTuple, arg: Union[int, str]) -> Any:
    if isinstance(arg, str):
        return getattr(self, arg)
    return self[arg]


def namedtuple_keys(self: NamedTuple) -> Iterable[str]:
    return self._fields


def namedtuple_values(self: NamedTuple) -> Iterable[Any]:
    return self


def namedtuple_items(self: NamedTuple) -> Iterable[Tuple[str, Any]]:
    return zip(self._fields, self)


def as_immutable_container(
    arg: Union[Tensor, Sequence, Mapping], recursive: bool = True
) -> Union[Tensor, tuple]:
    r"""Convert mutable container such as dict or list to an immutable container type.

    For use with ``torch.utils.tensorboard.SummaryWriter.add_graph`` when model output is list or dict.
    See error message: "Encountering a dict at the output of the tracer might cause the trace to be incorrect,
    this is only valid if the container structure does not change based on the module's inputs. Consider using
    a constant container instead (e.g. for `list`, use a `tuple` instead. for `dict`, use a `NamedTuple` instead).
    If you absolutely need this and know the side effects, pass strict=False to trace() to allow this behavior."

    """
    if recursive:
        if isinstance(arg, Mapping):
            arg = {key: as_immutable_container(value) for key, value in arg.items()}
        elif isinstance(arg, Sequence):
            arg = [as_immutable_container(value) for value in arg]
    if isinstance(arg, Mapping):
        output_type = namedtuple("Dict", sorted(arg.keys()))
        output_type.__getitem__ = get_namedtuple_item
        output_type.keys = namedtuple_keys
        output_type.values = namedtuple_values
        output_type.items = namedtuple_items
        return output_type(**arg)
    if isinstance(arg, list):
        return tuple(arg)
    return arg


def conv_output_size(
    in_size: ScalarOrTuple[int],
    kernel_size: ScalarOrTuple[int],
    stride: ScalarOrTuple[int] = 1,
    padding: ScalarOrTuple[int] = 0,
    dilation: ScalarOrTuple[int] = 1,
) -> ScalarOrTuple[int]:
    r"""Calculate spatial size of output tensor after convolution."""
    device = torch.device("cpu")
    m: Tensor = torch.atleast_1d(torch.tensor(in_size, dtype=torch.int, device=device))
    k: Tensor = torch.atleast_1d(torch.tensor(kernel_size, dtype=torch.int, device=device))
    s: Tensor = torch.atleast_1d(torch.tensor(stride, dtype=torch.int, device=device))
    d: Tensor = torch.atleast_1d(torch.tensor(dilation, dtype=torch.int, device=device))
    if m.ndim != 1:
        raise ValueError("conv_output_size() 'in_size' must be scalar or sequence")
    ndim = m.shape[0]
    if ndim == 1 and k.shape[0] > 1:
        ndim = k.shape[0]
    for arg, name in zip([k, s, d], ["kernel_size", "stride", "dilation"]):
        if arg.ndim != 1 or arg.shape[0] not in (1, ndim):
            raise ValueError(
                f"conv_output_size() {name!r} must be scalar or sequence of length {ndim}"
            )
    k = k.expand(ndim)
    s = s.expand(ndim)
    d = d.expand(ndim)
    if padding == "valid":
        padding = 0
    elif padding == "same":
        if not s.eq(1).all():
            raise ValueError("conv_output_size() padding='same' requires stride=1")
        padding = same_padding(kernel_size=kernel_size, dilation=dilation)
    elif isinstance(padding, str):
        raise ValueError("conv_output_size() 'padding' string must be 'valid' or 'same'")
    p: Tensor = torch.atleast_1d(torch.tensor(padding, dtype=torch.int, device=device))
    if p.ndim != 1 or p.shape[0] not in (1, ndim):
        raise ValueError(
            f"conv_output_size() 'padding' must be scalar or sequence of length {ndim}"
        )
    p = p.expand(ndim)
    n = p.mul(2).add_(m).sub_(k.sub(1).mul_(d)).sub_(1).float().div_(s).add_(1).floor_().long()
    if isinstance(in_size, int):
        return n[0].item()
    return Size(n.tolist())


def conv_transposed_output_size(
    in_size: ScalarOrTuple[int],
    kernel_size: ScalarOrTuple[int],
    stride: ScalarOrTuple[int] = 1,
    padding: ScalarOrTuple[int] = 0,
    output_padding: ScalarOrTuple[int] = 0,
    dilation: ScalarOrTuple[int] = 1,
) -> ScalarOrTuple[int]:
    r"""Calculate spatial size of output tensor after transposed convolution."""
    device = torch.device("cpu")
    m: Tensor = torch.atleast_1d(torch.tensor(in_size, dtype=torch.int, device=device))
    k: Tensor = torch.atleast_1d(torch.tensor(kernel_size, dtype=torch.int, device=device))
    s: Tensor = torch.atleast_1d(torch.tensor(stride, dtype=torch.int, device=device))
    p: Tensor = torch.atleast_1d(torch.tensor(padding, dtype=torch.int, device=device))
    o: Tensor = torch.atleast_1d(torch.tensor(output_padding, dtype=torch.int, device=device))
    d: Tensor = torch.atleast_1d(torch.tensor(dilation, dtype=torch.int, device=device))
    if m.ndim != 1:
        raise ValueError("conv_transposed_output_size() 'in_size' must be scalar or sequence")
    ndim = m.shape[0]
    if ndim == 1 and k.shape[0] > 1:
        ndim = k.shape[0]
    for arg, name in zip(
        [k, s, p, o, d],
        ["kernel_size", "stride", "padding", "output_padding", "dilation"],
    ):
        if arg.ndim != 1 or arg.shape[0] not in (1, ndim):
            raise ValueError(
                f"conv_transposed_output_size() {name!r} must be scalar or sequence of length {ndim}"
            )
    k = k.expand(ndim)
    s = s.expand(ndim)
    p = p.expand(ndim)
    o = o.expand(ndim)
    d = d.expand(ndim)
    n = m.sub(1).mul_(s).sub_(p.mul(2)).add_(k.sub(1).mul_(d)).add_(o).add_(1)
    if isinstance(in_size, int):
        return n.item()
    return Size(n.tolist())


def pad_output_size(
    in_size: ScalarOrTuple[int],
    padding: ScalarOrTuple[int] = 0,
) -> ScalarOrTuple[int]:
    r"""Calculate spatial size of output tensor after padding."""
    device = torch.device("cpu")
    m: Tensor = torch.atleast_1d(torch.tensor(in_size, dtype=torch.int, device=device))
    p: Tensor = torch.atleast_1d(torch.tensor(padding, dtype=torch.int, device=device))
    if m.ndim != 1:
        raise ValueError("pad_output_size() 'in_size' must be scalar or sequence")
    ndim = m.shape[0]
    if ndim == 1 and p.shape[0] > 1 and p.shape[0] % 2:
        ndim = p.shape[0] // 2
    if p.ndim != 1 or p.shape[0] not in (1, 2 * ndim):
        raise ValueError(
            f"pad_output_size() 'padding' must be scalar or sequence of length {2 * ndim}"
        )
    p = p.expand(2 * ndim)
    n = p.reshape(ndim, 2).sum(dim=1).add(m)
    if isinstance(in_size, int):
        return n[0].item()
    return Size(n.tolist())


def pool_output_size(
    in_size: ScalarOrTuple[int],
    kernel_size: ScalarOrTuple[int],
    stride: ScalarOrTuple[int] = 1,
    padding: ScalarOrTuple[int] = 0,
    dilation: ScalarOrTuple[int] = 1,
    ceil_mode: bool = False,
) -> ScalarOrTuple[int]:
    r"""Calculate spatial size of output tensor after pooling."""
    device = torch.device("cpu")
    m: Tensor = torch.atleast_1d(torch.tensor(in_size, dtype=torch.int, device=device))
    k: Tensor = torch.atleast_1d(torch.tensor(kernel_size, dtype=torch.int, device=device))
    s: Tensor = torch.atleast_1d(torch.tensor(stride, dtype=torch.int, device=device))
    p: Tensor = torch.atleast_1d(torch.tensor(padding, dtype=torch.int, device=device))
    d: Tensor = torch.atleast_1d(torch.tensor(dilation, dtype=torch.int, device=device))
    if m.ndim != 1:
        raise ValueError("pool_output_size() 'in_size' must be scalar or sequence")
    ndim = m.shape[0]
    if ndim == 1 and k.shape[0] > 1:
        ndim = k.shape[0]
    for arg, name in zip([k, s, p, d], ["kernel_size", "stride", "padding", "dilation"]):
        if arg.ndim != 1 or arg.shape[0] not in (1, ndim):
            raise ValueError(
                f"pool_output_size() {name!r} must be scalar or sequence of length {ndim}"
            )
    k = k.expand(ndim)
    s = s.expand(ndim)
    p = p.expand(ndim)
    d = d.expand(ndim)
    n = p.mul(2).add_(m).sub_(k.sub(1).mul_(d)).sub_(1).float().div_(s).add_(1)
    n = n.ceil() if ceil_mode else n.floor()
    n = n.long()
    if isinstance(in_size, int):
        return n[0].item()
    return Size(n.tolist())


def unpool_output_size(
    in_size: ScalarOrTuple[int],
    kernel_size: ScalarOrTuple[int],
    stride: ScalarOrTuple[int] = 1,
    padding: ScalarOrTuple[int] = 0,
) -> ScalarOrTuple[int]:
    r"""Calculate spatial size of output tensor after unpooling."""
    device = torch.device("cpu")
    m: Tensor = torch.atleast_1d(torch.tensor(in_size, dtype=torch.int, device=device))
    k: Tensor = torch.atleast_1d(torch.tensor(kernel_size, dtype=torch.int, device=device))
    s: Tensor = torch.atleast_1d(torch.tensor(stride, dtype=torch.int, device=device))
    p: Tensor = torch.atleast_1d(torch.tensor(padding, dtype=torch.int, device=device))
    if m.ndim != 1:
        raise ValueError("unpool_output_size() 'in_size' must be scalar or sequence")
    ndim = m.shape[0]
    if ndim == 1 and k.shape[0] > 1:
        ndim = k.shape[0]
    for arg, name in zip([k, s, p], ["kernel_size", "stride", "padding"]):
        if arg.ndim != 1 or arg.shape[0] not in (1, ndim):
            raise ValueError(
                f"unpool_output_size() {name!r} must be scalar or sequence of length {ndim}"
            )
    k = k.expand(ndim)
    s = s.expand(ndim)
    p = p.expand(ndim)
    n = m.sub(1).mul_(s).sub_(p.mul(2)).add(k)
    if isinstance(in_size, int):
        return n[0].item()
    return Size(n.tolist())


@overload
def same_padding(kernel_size: int, dilation: int = 1) -> int:
    ...


@overload
def same_padding(kernel_size: Tuple[int, ...], dilation: int = 1) -> Tuple[int, ...]:
    ...


@overload
def same_padding(kernel_size: int, dilation: Tuple[int, ...]) -> Tuple[int, ...]:
    ...


@overload
def same_padding(kernel_size: Tuple[int, ...], dilation: Tuple[int, ...]) -> Tuple[int, ...]:
    ...


def same_padding(
    kernel_size: ScalarOrTuple[int], dilation: ScalarOrTuple[int] = 1
) -> ScalarOrTuple[int]:
    r"""Padding value needed to ensure convolution preserves input tensor shape.

    Return the padding value needed to ensure a convolution using the given kernel size produces an output of the same
    shape as the input for a stride of 1, otherwise ensure a shape of the input divided by the stride rounded down.

    Raises:
        NotImplementedError: When ``(kernel_size - 1) * dilation`` is an odd number.

    """
    # Adapted from Project MONAI
    # https://github.com/Project-MONAI/MONAI/blob/db8f7877da06a9b3710071c626c0488676716be1/monai/networks/layers/convutils.py
    device = torch.device("cpu")
    k: Tensor = torch.atleast_1d(torch.tensor(kernel_size, dtype=torch.int, device=device))
    d: Tensor = torch.atleast_1d(torch.tensor(dilation, dtype=torch.int, device=device))
    if k.ndim != 1:
        raise ValueError("same_padding() 'kernel_size' must be scalar or sequence")
    ndim = k.shape[0]
    if ndim == 1 and d.shape[0] > 1:
        ndim = d.shape[0]
    for arg, name in zip([k, d], ["kernel_size", "dilation"]):
        if arg.ndim != 1 or arg.shape[0] not in (1, ndim):
            raise ValueError(f"same_padding() {name!r} must be scalar or sequence of length {ndim}")
    if k.sub(1).mul(d).fmod(2).eq(1).any():
        raise NotImplementedError(
            f"Same padding not available for kernel_size={tuple(k.tolist())} and dilation={tuple(d.tolist())}."
        )
    p = k.sub(1).div(2).mul(d).type(torch.int)
    if isinstance(kernel_size, int) and isinstance(dilation, int):
        return p[0].item()
    return tuple(p.tolist())


@overload
def stride_minus_kernel_padding(kernel_size: int, stride: int) -> int:
    ...


@overload
def stride_minus_kernel_padding(kernel_size: Sequence[int], stride: int) -> Tuple[int, ...]:
    ...


@overload
def stride_minus_kernel_padding(kernel_size: int, stride: Sequence[int]) -> Tuple[int, ...]:
    ...


def stride_minus_kernel_padding(
    kernel_size: ScalarOrTuple[int], stride: ScalarOrTuple[int]
) -> ScalarOrTuple[int]:
    # Adapted from Project MONAI
    # https://github.com/Project-MONAI/MONAI/blob/db8f7877da06a9b3710071c626c0488676716be1/monai/networks/layers/convutils.py
    device = torch.device("cpu")
    k: Tensor = torch.atleast_1d(torch.tensor(kernel_size, dtype=torch.int, device=device))
    s: Tensor = torch.atleast_1d(torch.tensor(stride, dtype=torch.int, device=device))
    if k.ndim != 1:
        raise ValueError("stride_minus_kernel_padding() 'kernel_size' must be scalar or sequence")
    ndim = k.shape[0]
    if ndim == 1 and s.shape[0] > 1:
        ndim = s.shape[0]
    for arg, name in zip([k, s], ["kernel_size", "stride"]):
        if arg.ndim != 1 or arg.shape[0] not in (1, ndim):
            raise ValueError(
                f"stride_minus_kernel_padding() {name!r} must be scalar or sequence of length {ndim}"
            )
    assert k.ndim == 1, "stride_minus_kernel_padding() 'kernel_size' must be scalar or sequence"
    assert s.ndim == 1, "stride_minus_kernel_padding() 'stride' must be scalar or sequence"
    p = s.sub(k).type(torch.int)
    if isinstance(kernel_size, int) and isinstance(stride, int):
        return p[0].item()
    return tuple(p.tolist())


def upsample_padding(
    kernel_size: ScalarOrTuple[int], scale_factor: ScalarOrTuple[int]
) -> Tuple[int, ...]:
    r"""Padding on both sides for transposed convolution."""
    device = torch.device("cpu")
    k: Tensor = torch.atleast_1d(torch.tensor(kernel_size, dtype=torch.int, device=device))
    s: Tensor = torch.atleast_1d(torch.tensor(scale_factor, dtype=torch.int, device=device))
    assert k.ndim == 1, "upsample_padding() 'kernel_size' must be scalar or sequence"
    assert s.ndim == 1, "upsample_padding() 'scale_factor' must be scalar or sequence"
    p = k.sub(s).add(1).div(2).type(torch.int)
    if p.lt(0).any():
        raise ValueError(
            "upsample_padding() 'kernel_size' must be greater than or equal to 'scale_factor'"
        )
    return tuple(p.tolist())


def upsample_output_padding(
    kernel_size: ScalarOrTuple[int],
    scale_factor: ScalarOrTuple[int],
    padding: ScalarOrTuple[int],
) -> Tuple[int, ...]:
    r"""Output padding on one side for transposed convolution."""
    device = torch.device("cpu")
    k: Tensor = torch.atleast_1d(torch.tensor(kernel_size, dtype=torch.int, device=device))
    s: Tensor = torch.atleast_1d(torch.tensor(scale_factor, dtype=torch.int, device=device))
    p: Tensor = torch.atleast_1d(torch.tensor(padding, dtype=torch.int, device=device))
    assert k.ndim == 1, "upsample_output_padding() 'kernel_size' must be scalar or sequence"
    assert s.ndim == 1, "upsample_output_padding() 'scale_factor' must be scalar or sequence"
    assert p.ndim == 1, "upsample_output_padding() 'padding' must be scalar or sequence"
    op = p.mul(2).sub(k).add(s).type(torch.int)
    if op.lt(0).any():
        raise ValueError(
            "upsample_output_padding() 'output_padding' must be greater than or equal to zero"
        )
    return tuple(op.tolist())


def upsample_output_size(
    in_size: ScalarOrTuple[int],
    size: Optional[ScalarOrTuple[int]] = None,
    scale_factor: Optional[ScalarOrTuple[float]] = None,
) -> ScalarOrTuple[int]:
    r"""Calculate spatial size of output tensor after unpooling."""
    if size is not None and scale_factor is not None:
        raise ValueError("upsample_output_size() 'size' and 'scale_factor' are mutually exclusive")
    device = torch.device("cpu")
    m: Tensor = torch.atleast_1d(torch.tensor(in_size, dtype=torch.int, device=device))
    if m.ndim != 1:
        raise ValueError("upsample_output_size() 'in_size' must be scalar or sequence")
    ndim = m.shape[0]
    if size is not None:
        s: Tensor = torch.atleast_1d(torch.tensor(size, dtype=torch.int, device=device))
        if s.ndim != 1 or s.shape[0] not in (1, ndim):
            raise ValueError(
                f"upsample_output_size() 'size' must be scalar or sequence of length {ndim}"
            )
        n = s.expand(ndim)
    elif scale_factor is not None:
        s: Tensor = torch.atleast_1d(torch.tensor(scale_factor, dtype=torch.int, device=device))
        if s.ndim != 1 or s.shape[0] not in (1, ndim):
            raise ValueError(
                f"upsample_output_size() 'scale_factor' must be scalar or sequence of length {ndim}"
            )
        n = m.float().mul(s).floor().long()
    else:
        n = m
    if isinstance(in_size, int):
        return n.item()
    return Size(n.tolist())



# Upsample and SubpixelUpsample layers are adapted from MONAI project
# (https://github.com/Project-MONAI/MONAI/blob/db8f7877da06a9b3710071c626c0488676716be1/monai/networks/blocks/upsample.py)
#
# Copyright 2020 - 2021 MONAI Consortium
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Optional, Sequence, Union

import torch
from torch import Tensor, nn
from torch.nn import Module, Sequential, Identity

from deepali.core.enum import PaddingMode, Sampling
from deepali.core.grid import ALIGN_CORNERS
from deepali.core.nnutils import upsample_padding, upsample_output_padding
from deepali.core.typing import ScalarOrTuple
from deepali.modules import Pad

from .conv import convolution
from .pool import pooling


__all__ = ("Upsample", "UpsampleMode", "SubpixelUpsample")


class UpsampleMode(Enum):
    DECONV = "deconv"
    INTERPOLATE = "interpolate"
    PIXELSHUFFLE = "pixelshuffle"


class Upsample(Sequential):
    r"""Upsamples data by `scale_factor`.

    This module is adapted from the MONAI project. Supported modes are:

        - "deconv": uses a transposed convolution.
        - "interpolate": uses :py:class:`torch.nn.Upsample`.
        - "pixelshuffle": uses :py:class:`monai.networks.blocks.SubpixelUpsample`.

    This module can optionally apply a convolution prior to upsampling interpolation
    (e.g., used to map the number of features from `in_channels` to `out_channels`).

    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: Optional[int] = None,
        out_channels: Optional[int] = None,
        scale_factor: Union[Sequence[Union[int, float]], Union[int, float]] = 2,
        kernel_size: Optional[ScalarOrTuple[int]] = None,
        mode: Union[UpsampleMode, str] = "default",
        pre_conv: Optional[Union[Module, str, int]] = "default",
        sampling: Union[Sampling, str] = Sampling.LINEAR,
        align_corners: bool = ALIGN_CORNERS,
        apply_pad_pool: bool = True,
        padding_mode: Union[PaddingMode, str] = PaddingMode.ZEROS,
        bias: bool = True,
        init: str = "default",
    ) -> None:
        r"""Initialize upsampling layer.

        Args:
            spatial_dims: Number of spatial dimensions of input tensor.
            in_channels: Number of channels of the input tensor.
            out_channels: Number of channels of the output tensor. Defaults to `in_channels`.
            scale_factor: Multiplier for spatial size. Has to match input size if it is a tuple.
            kernel_size: Kernel size of transposed convolution in "deconv" mode or default ``pre_conv`` otherwise.
                The default kernel size in "deconv" mode is equal to the specified ``scale_factor``. The default
                ``pre_conv`` kernel size in "interpolate" mode is 1. In "pixelshuffle" mode, the default is 3.
            mode: Upsampling mode: "deconv", "interpolate", or "pixelshuffle".
            pre_conv: A conv block applied before upsampling. When ``conv_block`` is ``"default"``, one reserved
                conv layer will be utilized. This argument is only used for "interpolate" or "pixelshuffle" mode.
            sampling: Interpolation mode used for ``torch.nn.Upsample`` in "interpolate" mode.
            align_corners: See `torch.nn.Upsample`.
            apply_pad_pool: If True the upsampled tensor is padded then average pooling is applied with a kernel the
                size of `scale_factor` with a stride of 1. Only used in the pixelshuffle mode.
            padding_mode: Padding mode to use for default ``pre_conv`` and pixelshuffle ``apply_pad_pool``.
            bias: Whether to have a bias term in the default preconv and deconv layers.
            init: How to initialize default ``conv_block`` weights (cf. ``convolution()``). In case of "pixelshuffle" mode,
                if value is "icnr" or "default", ICNR initialization is used. Use "uniform" for initialization without ICNR.

        """
        super().__init__()

        if out_channels is None:
            out_channels = in_channels
        upsample_mode = UpsampleMode.DECONV if mode == "default" else UpsampleMode(mode)
        padding_mode = PaddingMode(padding_mode).conv_mode(spatial_dims)

        if upsample_mode == UpsampleMode.DECONV:
            if not in_channels:
                raise ValueError(
                    f"{type(self).__name__}() 'in_channels' required in {upsample_mode.value!r} mode"
                )

            if isinstance(scale_factor, (int, float)):
                scale_factor = (scale_factor,) * spatial_dims
            elif len(scale_factor) != spatial_dims:
                raise ValueError(
                    f"{type(self).__name__}() 'scale_factor' must be scalar or sequence of length {spatial_dims}"
                )
            try:
                scale_factor = tuple(int(s) for s in scale_factor)
            except TypeError:
                raise TypeError(
                    f"{type(self).__name__}() 'scale_factor' must be ints for {upsample_mode.value!r} mode"
                )

            if kernel_size is None:
                kernel_size = scale_factor
            elif isinstance(kernel_size, (int, float)):
                kernel_size = (kernel_size,) * spatial_dims
            elif len(kernel_size) != spatial_dims:
                raise ValueError(
                    f"{type(self).__name__}() 'kernel_size' must be scalar or sequence of length {spatial_dims}"
                )
            if any(k < s for k, s in zip(kernel_size, scale_factor)):
                raise ValueError(
                    f"{type(self).__name__}() 'kernel_size' must be greater than or equal to 'scale_factor'"
                )

            # Output size should be scale factor times input size, thus:
            # 2 * padding - output_padding == kernel_size - scale_factor
            padding = upsample_padding(kernel_size, scale_factor)
            output_padding = upsample_output_padding(kernel_size, scale_factor, padding)

            deconv = convolution(
                spatial_dims=spatial_dims,
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=scale_factor,
                padding=padding,
                output_padding=output_padding,
                dilation=1,
                init=init,
                bias=bias,
                transposed=True,
            )
            self.add_module("deconv", deconv)

        elif upsample_mode == UpsampleMode.INTERPOLATE:
            if pre_conv == "default":
                if in_channels is None or in_channels == out_channels:
                    pre_conv = None
                else:
                    if kernel_size is None:
                        kernel_size = 1
                    if isinstance(kernel_size, int):
                        kernel_size = (kernel_size,) * spatial_dims
                    elif not isinstance(kernel_size, Sequence):
                        raise TypeError(
                            f"{type(self).__name__}() 'kernel_size' must be int or Sequence[int]"
                        )
                    elif len(kernel_size) != spatial_dims:
                        raise ValueError(
                            f"{type(self).__name__}() 'kernel_size' must be int or {spatial_dims}-tuple"
                        )
                    if any(k < 1 for k in kernel_size):
                        raise ValueError(f"{type(self).__name__}() 'kernel_size' must be positive")
                    if any(ks % 2 == 0 for ks in kernel_size):
                        padding = tuple(((ks - 1) // 2, ks // 2) for ks in kernel_size)
                        padding = tuple(p for a, b in reversed(padding) for p in (a, b))
                        pre_pad = Pad(padding=padding, mode=padding_mode)
                        self.add_module("prepad", pre_pad)
                        padding = 0
                    else:
                        padding = tuple(ks // 2 for ks in kernel_size)
                    pre_conv = convolution(
                        spatial_dims=spatial_dims,
                        in_channels=in_channels,
                        out_channels=out_channels,
                        kernel_size=kernel_size,
                        padding=padding,
                        padding_mode=padding_mode,
                        init=init,
                        bias=bias,
                    )
            if pre_conv is not None:
                if not isinstance(pre_conv, Module):
                    raise TypeError(
                        f"{type(self).__name__}() 'preconv' must be string 'default' or Module"
                    )
                self.add_module("preconv", pre_conv)

            mode = Sampling(sampling).interpolate_mode(spatial_dims)
            upsample = nn.Upsample(
                scale_factor=scale_factor, mode=mode, align_corners=align_corners
            )
            self.add_module("interpolate", upsample)

        elif upsample_mode == UpsampleMode.PIXELSHUFFLE:
            try:
                scale_factor_ = int(scale_factor)
            except TypeError:
                raise TypeError(
                    f"{type(self).__name__}() 'scale_factor' must be int for {upsample_mode.value!r} mode"
                )

            if kernel_size is None:
                kernel_size = 3

            module = SubpixelUpsample(
                spatial_dims=spatial_dims,
                in_channels=in_channels,
                out_channels=out_channels,
                scale_factor=scale_factor_,
                conv_block=pre_conv,
                apply_pad_pool=apply_pad_pool,
                kernel_size=kernel_size,
                padding_mode=padding_mode,
                init=init,
                bias=bias,
            )
            self.add_module("pixelshuffle", module)

        else:
            raise NotImplementedError(f"{type(self).__name__}() mode={mode!r} not implemented")


class SubpixelUpsample(Module):
    r"""Upsample using a subpixel CNN.

    This module is adapted from the MONAI project and supports 1D, 2D and 3D input images.
    The module consists of two parts. First of all, a convolutional layer is employed
    to increase the number of channels into: ``in_channels * (scale_factor ** spatial_dims)``.
    Secondly, a pixel shuffle manipulation is utilized to aggregate the feature maps from
    low resolution space and build the super resolution space. The first part of the module
    is not fixed, a sequential layer can be used to replace the default single layer.

    See: Shi et al., 2016, "Real-Time Single Image and Video Super-Resolution
    Using a nEfficient Sub-Pixel Convolutional Neural Network."

    See: Aitken et al., 2017, "Checkerboard artifact free sub-pixel convolution".

    The idea comes from:
    https://arxiv.org/abs/1609.05158

    The pixel shuffle mechanism refers to:
    https://github.com/pytorch/pytorch/blob/master/aten/src/ATen/native/PixelShuffle.cpp
    and:
    https://github.com/pytorch/pytorch/pull/6340/files

    """

    def __init__(
        self,
        spatial_dims: int,
        in_channels: Optional[int],
        out_channels: Optional[int] = None,
        scale_factor: Union[int, float] = 2,
        conv_block: Optional[Union[nn.Module, str, int]] = "default",
        apply_pad_pool: bool = True,
        kernel_size: ScalarOrTuple[int] = 3,
        padding_mode: Union[PaddingMode, str] = PaddingMode.ZEROS,
        init: str = "default",
        bias: Union[bool, str] = True,
    ) -> None:
        r"""Initialize upsampling layer.

        Args:
            spatial_dims: Number of spatial dimensions of the input image.
            in_channels: Number of channels of the input image.
            out_channels: Optional number of channels of the output image.
            scale_factor: Multiplier for spatial size. Must be castable to ``int``. Defaults to 2.
            conv_block: A conv block to extract feature maps before upsampling.

                - When ``"default"``, one reserved conv layer will be utilized.
                - When ``torch..nn.Module``, the output number of channels must be divisible by ``(scale_factor ** spatial_dims)``.

            apply_pad_pool: If True the upsampled tensor is padded then average pooling is applied with a kernel the
                size of `scale_factor` with a stride of 1. This implements the nearest neighbour resize convolution
                component of subpixel convolutions described in Aitken et al.
            kernel_size: Size of default ``conv_block`` kernel. Defaults to 3.
            padding_mode: Padding mode to use for default ``conv_block`` and ``apply_pad_pool``.
            init: How to initialize default ``conv_block`` weights (cf. ``convolution()``). If value is "icnr" or
                "default", ICNR initialization is used. Use "uniform" for standard initialization without ICNR.
            bias: Whether to have a bias term in the default conv_block. When a string is given, it specifies how
                the bias term is initialized (cf. ``convolution()``).

        """
        super().__init__()

        try:
            scale_factor = int(scale_factor)
        except TypeError:
            raise TypeError("SubpixelUpsample() 'scale_factor' must be int")
        if scale_factor < 1:
            raise ValueError("SubpixelUpsample() 'scale_factor' must be a positive integer")

        if init in ("icnr", "ICNR"):
            init = "default"

        self.spatial_dims = spatial_dims
        self.scale_factor = scale_factor

        if conv_block == "default":
            if not in_channels:
                raise ValueError("SubpixelUpsample() 'in_channels' required")
            out_channels = out_channels or in_channels
            conv_out_channels = out_channels * (scale_factor**spatial_dims)
            if kernel_size % 2 == 0:
                padding = ((kernel_size - 1) // 2, kernel_size // 2) * spatial_dims
                pre_pad = Pad(padding=padding, mode=padding_mode)
                padding = 0
            else:
                pre_pad = None
                padding = kernel_size // 2
            conv_block = convolution(
                spatial_dims=spatial_dims,
                in_channels=in_channels,
                out_channels=conv_out_channels,
                kernel_size=kernel_size,
                stride=1,
                padding=padding,
                padding_mode=padding_mode,
                init=init,
                bias=bias,
            )
            if init == "default":
                icnr_init(conv_block.weight, scale_factor)
            if pre_pad is not None:
                conv_block = Sequential(pre_pad, conv_block)
        elif conv_block is None:
            conv_block = Identity()
        elif not isinstance(conv_block, Module):
            raise ValueError(
                "SubpixelUpsample() 'conv_block' must be string 'default', Module, or None"
            )
        self.conv_block = conv_block

        if apply_pad_pool:
            pad_pool = Sequential(
                Pad(
                    padding=(scale_factor - 1, 0) * spatial_dims,
                    mode=padding_mode,
                    value=0,
                ),
                pooling("avg", spatial_dims=spatial_dims, kernel_size=scale_factor, stride=1),
            )
        else:
            pad_pool = Identity()
        self.pad_pool = pad_pool

    def forward(self, x: Tensor) -> Tensor:
        r"""

        Args:
            x: Tensor in shape (batch, channel, spatial_1[, spatial_2, ...).

        """
        x = self.conv_block(x)
        x = pixelshuffle(x, self.spatial_dims, self.scale_factor)
        x = self.pad_pool(x)
        return x


def icnr_init(weight: Tensor, upsample_factor: int, init=nn.init.kaiming_normal_) -> None:
    r"""ICNR initialization for 2D/3D kernels.

    Adapted from MONAI project and based on Aitken et al., 2017, "Checkerboard artifact free sub-pixel convolution".

    """
    out_channels, in_channels, *dims = weight.shape
    scale_factor = upsample_factor ** len(dims)

    oc2 = int(out_channels / scale_factor)

    kernel = torch.zeros([oc2, in_channels] + dims)
    kernel: Tensor = init(kernel)
    kernel = kernel.transpose(0, 1)
    kernel = kernel.reshape(oc2, in_channels, -1)
    kernel = kernel.repeat(1, 1, scale_factor)
    kernel = kernel.reshape([in_channels, out_channels] + dims)
    kernel = kernel.transpose(0, 1)
    weight.data.copy_(kernel)


def pixelshuffle(x: Tensor, spatial_dims: int, scale_factor: int) -> Tensor:
    r"""Apply pixel shuffle to the tensor `x` with spatial dimensions `spatial_dims` and scaling factor `scale_factor`.

    See: Shi et al., 2016, "Real-Time Single Image and Video Super-Resolution
    Using an Efficient Sub-Pixel Convolutional Neural Network."

    See: Aitken et al., 2017, "Checkerboard artifact free sub-pixel convolution".

    Args:
        x: Input tensor
        spatial_dims: number of spatial dimensions, typically 2 or 3 for 2D or 3D
        scale_factor: factor to rescale the spatial dimensions by, must be >=1

    Returns:
        Reshuffled version of `x`.

    Raises:
        ValueError: When input channels of `x` are not divisible by (scale_factor ** spatial_dims)

    """

    dim, factor = spatial_dims, scale_factor
    input_size = list(x.size())
    batch_size, channels = input_size[:2]
    scale_divisor = factor**dim

    if channels % scale_divisor != 0:
        raise ValueError(
            f"pixelshuffle() number of input channels ({channels}) must be evenly"
            f" divisible by scale_factor ** spatial_dims ({factor}**{dim}={scale_divisor})"
        )

    org_channels = channels // scale_divisor
    output_size = [batch_size, org_channels] + [d * factor for d in input_size[2:]]

    indices = tuple(range(2, 2 + 2 * dim))
    indices_factor, indices_dim = indices[:dim], indices[dim:]
    permute_indices = (0, 1) + sum(zip(indices_dim, indices_factor), ())

    x = x.reshape(batch_size, org_channels, *([factor] * dim + input_size[2:]))
    x = x.permute(permute_indices).reshape(output_size)
    return x



r"""Bridge between VTK and SimpleITK."""

from typing import Optional, Sequence, Union

import SimpleITK as sitk
from vtk import vtkImageData, vtkImageImport, vtkPoints, vtkPointSet
from vtk import (
    VTK_CHAR,
    VTK_SHORT,
    VTK_LONG,
    VTK_LONG_LONG,
    VTK_UNSIGNED_CHAR,
    VTK_UNSIGNED_SHORT,
    VTK_UNSIGNED_LONG,
    VTK_UNSIGNED_LONG_LONG,
    VTK_FLOAT,
    VTK_DOUBLE,
)

from deepali.utils.simpleitk.grid import GridAttrs
from deepali.utils.simpleitk.sample import interpolate_ndimage

from .numpy import numpy_to_vtk_array, vtk_to_numpy_array, vtk_to_numpy_points


VTK_DATA_TYPE_FROM_SITK_PIXEL_ID = {
    sitk.sitkInt8: VTK_CHAR,
    sitk.sitkInt16: VTK_SHORT,
    sitk.sitkInt32: VTK_LONG,
    sitk.sitkInt64: VTK_LONG_LONG,
    sitk.sitkUInt8: VTK_UNSIGNED_CHAR,
    sitk.sitkUInt16: VTK_UNSIGNED_SHORT,
    sitk.sitkUInt32: VTK_UNSIGNED_LONG,
    sitk.sitkUInt64: VTK_UNSIGNED_LONG_LONG,
    sitk.sitkFloat32: VTK_FLOAT,
    sitk.sitkFloat64: VTK_DOUBLE,
    sitk.sitkVectorInt8: VTK_CHAR,
    sitk.sitkVectorInt16: VTK_SHORT,
    sitk.sitkVectorInt32: VTK_LONG,
    sitk.sitkVectorInt64: VTK_LONG_LONG,
    sitk.sitkVectorUInt8: VTK_UNSIGNED_CHAR,
    sitk.sitkVectorUInt16: VTK_UNSIGNED_SHORT,
    sitk.sitkVectorUInt32: VTK_UNSIGNED_LONG,
    sitk.sitkVectorUInt64: VTK_UNSIGNED_LONG_LONG,
    sitk.sitkVectorFloat32: VTK_FLOAT,
    sitk.sitkVectorFloat64: VTK_DOUBLE,
}


def apply_warp_field_to_points(
    warp_field: sitk.Image, points: vtkPoints, is_def_field: bool = False
) -> vtkPoints:
    r"""Transform vtkPoints by linearly interpolated dense vector field.

    Args:
        warp_field: Vector field that acts on the given points.
            If ``None``, an identity deformation is assumed.
        points: Input points in physical space of warp field.
        is_def_field: If ``True``, the input ``warp_field`` must be a vector field
            of output coordinates. Otherwise, ``warp_field`` must contain displacements
            in physical image space.

    Returns:
        Transformed output points.

    """
    out = points.NewInstance()

    if warp_field is None:
        out.DeepCopy(points)
        return out

    x = vtk_to_numpy_points(points)
    y = interpolate_ndimage(warp_field, x)
    if not is_def_field:
        y += x
    out.SetData(numpy_to_vtk_array(y))
    return out


def apply_warp_field_to_pointset(
    warp_field: sitk.Image, pointset: vtkPointSet, is_def_field: bool = False
) -> vtkPointSet:
    r"""Transform vtkPoints of vtkPointSet by linearly interpolated dense displacement field.

    Args:
        warp_field: Vector field that acts on the given points.
            If ``None``, an identity deformation is assumed.
        pointset: Input point set (e.g., vtkPolyData surface mesh).
        is_def_field: If ``True``, the input ``warp_field`` must be a vector field
            of output coordinates. Otherwise, ``warp_field`` must contain displacements
            in physical image space.

    Returns:
        Deep copy of input ``pointset`` with transformed points.

    """
    output = pointset.NewInstance()
    output.DeepCopy(pointset)
    if warp_field is not None:
        points = apply_warp_field_to_points(
            warp_field, pointset.GetPoints(), is_def_field=is_def_field
        )
        output.SetPoints(points)
    return output


def image_data_grid(data: vtkImageData) -> GridAttrs:
    r"""Create image grid from vtkImageData object."""
    extent = data.GetExtent()
    return GridAttrs(
        size=(
            extent[1] - extent[0] + 1,
            extent[3] - extent[2] + 1,
            extent[5] - extent[4] + 1,
        ),
        origin=data.GetOrigin(),
        spacing=data.GetSpacing(),
    )


def vtk_image_from_sitk_image(
    image: sitk.Image, spacing: Optional[Union[float, Sequence[float]]] = None
) -> vtkImageData:
    r"""Create vtkImageData from SimpleITK image.

    VTK versions before 9.0 do not support orientation information. Therefore, use only the pixel spacing
    information for index to physical space transformations when working with VTK data structures. These
    can be mapped into the original physical space by applying the rotation and translation from the
    original image grid.

    Args:
        image: SimpleITK image.
        spacing: Data spacing to use instead of ``image.GetSpacing()``.

    Returns:
        vtk_image: vtkImageData instance with spacing set, but origin equal to (0, 0, 0).

    """
    data = sitk.GetArrayFromImage(image).tobytes()
    pixel_id = image.GetPixelIDValue()
    data_type = VTK_DATA_TYPE_FROM_SITK_PIXEL_ID[pixel_id]
    size = list(image.GetSize())
    if spacing is None:
        spacing = image.GetSpacing()
    elif isinstance(spacing, (int, float)):
        spacing = (spacing,) * len(size)
    spacing = list(spacing)
    while len(size) < 3:
        size.append(1)
        spacing.append(1.0)
    importer = vtkImageImport()
    importer.CopyImportVoidPointer(data, len(data))
    importer.SetDataScalarType(data_type)
    importer.SetNumberOfScalarComponents(image.GetNumberOfComponentsPerPixel())
    importer.SetDataSpacing(spacing)
    importer.SetWholeExtent(0, size[0] - 1, 0, size[1] - 1, 0, size[2] - 1)
    importer.SetDataExtentToWholeExtent()
    importer.UpdateWholeExtent()
    output = vtkImageData()
    output.DeepCopy(importer.GetOutput())
    return output


def sitk_image_from_vtk_image(image: vtkImageData, grid: Optional[GridAttrs] = None) -> sitk.Image:
    r"""Create SimpleITK image from vtkImageData."""
    if image.GetNumberOfScalarComponents() != 1:
        raise NotImplementedError("sitk_image_from_vtk_image() only supports scalar 'image'")
    data = image.GetPointData().GetScalars()
    data = vtk_to_numpy_array(data)
    data = data.reshape(tuple(reversed(grid.size)))
    output = sitk.GetImageFromArray(data)
    output.SetOrigin(grid.origin)
    output.SetSpacing(grid.spacing)
    output.SetDirection(grid.direction)
    return output



from deprecation import deprecated

from deepali.core import __version__
from .transformer import ImageTransformer


@deprecated(
    deprecated_in="0.3",
    removed_in="1.0",
    current_version=__version__,
    details="Use deepali.spatial.ImageTransformer instead",
)
class ImageTransform(ImageTransformer):
    ...



r"""Reference objects stored in AWS Simple Storage Service (S3)."""

from __future__ import annotations

import re

from copy import deepcopy
from pathlib import Path
from tempfile import gettempdir
from typing import Generator, Optional

from deepali.core.pathlib import PathStr, to_uri
from deepali.core.storage import StorageObject

from .client import S3Client


class S3Object(StorageObject):
    r"""Object stored in AWS Simple Storage Service (S3)."""

    def __init__(self, bucket: str, key: str, path: PathStr = None) -> None:
        r"""Initialize AWS S3 object.

        Args:
            bucket: Name of AWS S3 bucket containing this object.
            key: Key of object in AWS S3 bucket with forward slashes as path separators.
                When string ends with a forward slash, the S3 object represents a set
                of S3 objects which share this common key prefix. The local path is
                in this case a directory tree rather than a single file path.
            path: File path of local file or directory corresponding to AWS S3 object.
                When not specified, a temporary path in the tempfile.gettempdir() is
                constructed from the bucket name and S3 object key.

        """
        if bucket is None:
            raise TypeError("S3 bucket name must be str")
        if bucket == "":
            raise ValueError("S3 bucket name must not be an empty string")
        key = self.normkey(key)
        path = Path(path) if path else self.default_path(bucket=bucket, key=key)
        if not key.endswith("/") and path.is_dir():
            key += "/"
        super().__init__(path)
        self._bucket = bucket
        self._key = key

    @property
    def s3(self) -> S3Client:
        r"""
        Returns:
            client: Underlying connected S3 client.

        """
        return S3Client.default().connect()

    @staticmethod
    def normkey(key: str) -> str:
        r"""Normalize S3 object key string."""
        if not isinstance(key, str):
            raise TypeError("S3 object key must be str")
        if key == "":
            return "/"
        key = re.sub(r"[/\\]{1,}", "/", key)
        if len(key) > 1 and key.startswith("/"):
            key = key[1:]
        return key

    @staticmethod
    def default_path(bucket: str, key: str) -> Path:
        r"""Get default local path of S3 object copy.

        Args:
            bucket: Name of S3 bucket.
            key: Key of S3 object.

        Returns:
            path: Absolute path of local copy of S3 object.

        """
        key = S3Object.normkey(key)
        if key.startswith("/"):
            key = key[1:]
        return Path(gettempdir()).joinpath("deepali", "data", "s3", bucket, key)

    def reset_path(self) -> S3Object:
        r"""Reset ``path`` to ``default_path`` for given ``bucket`` and ``key``.

        Returns:
            self: This instance.

        """
        self._path = self.default_path(bucket=self.bucket, key=self.key)
        return self

    @staticmethod
    def from_path(*args: Optional[PathStr]) -> S3Object:
        r"""Create storage object from path or URI.

        Args:
            args: Path or URI components. The last absolute path or URI is the base to which
                subsequent arguments are appended. Any ``None`` value is ignored. See also ``to_uri()``.

        Returns:
            obj (Resource): Instance of concrete type representing the referenced storage object.

        """
        return S3Object.from_uri(to_uri(*args))

    @classmethod
    def from_uri(cls, uri: str) -> S3Object:
        r"""Create AWS S3 object from URI.

        Args:
            uri: URI of S3 object. Must start with 's3://' followed by the bucket name.
                The remainder of the URI represents the object key, excluding the forward
                slash separating the bucket name from the object key.

        Returns:
            obj: S3 object instance.

        """
        match = re.match(r"[sS]3://(?P<bucket>[^/]+)/(?P<key>.*)", uri)
        if not match:
            raise ValueError("Invalid AWS S3 object URI: %s", uri)
        return cls(bucket=match["bucket"], key=match["key"])

    @property
    def uri(self) -> str:
        r"""URI of storage object."""
        return "s3://" + self.bucket + "/" + self.key

    @property
    def bucket(self) -> str:
        r"""Name of S3 bucket containing storage object."""
        return self._bucket

    @property
    def key(self) -> str:
        r"""Name of S3 key corresponding to storage object referenced by this instance."""
        return self._key

    @property
    def name(self) -> str:
        r"""Name of storage object including file name extension, excluding directory path."""
        key = self.key
        if key.endswith("/"):
            key = key[:-1]
        return key.rsplit("/", 1)[-1]

    def with_bucket(self, bucket: str) -> S3Object:
        r"""Create copy of storage object reference with modified ``bucket``.

        Args:
            bucket: New bucket name.

        Returns:
            self: New storage object reference with modified ``bucket`` property.

        """
        if bucket is None:
            raise TypeError("Bucket name must be str")
        if bucket == "":
            raise ValueError("Invalid S3 bucket name")
        obj = deepcopy(self)
        obj._bucket = bucket
        return obj

    def with_key(self, key: str) -> S3Object:
        r"""Create copy of storage object reference with modified ``key``.

        Args:
            key: New S3 object key.

        Returns:
            self: New storage object reference with modified ``key`` property.

        """
        if key is None:
            raise TypeError("S3 object key must be str")
        if key == "":
            raise ValueError("Invalid S3 object key")
        obj = deepcopy(self)
        obj._key = key
        return obj

    def exists(self) -> bool:
        r"""Whether object exists in AWS S3."""
        return self.s3.exists(bucket=self.bucket, key=self.key)

    def is_file(self) -> bool:
        r"""Whether AWS S3 object exists and represents a file."""
        return not self.key.endswith("/") and self.exists()

    def is_dir(self) -> bool:
        r"""Whether AWS S3 object exists and represents a directory."""
        return self.key.endswith("/") and self.exists()

    def iterdir(self, prefix: str = None) -> Generator[S3Object, None, None]:
        r"""List S3 objects within directory, excluding subfolder contents.

        Args:
            prefix: Name prefix.

        Returns:
            iterable: Generator of S3 objects.

        """
        if self.key.endswith("/"):
            if prefix is None:
                prefix = ""
            if self.key != "/":
                prefix = self.key + prefix
            for key in self.s3.iterdir(bucket=self.bucket, prefix=prefix):
                yield S3Object(bucket=self.bucket, key=key)

    def pull(self, force: bool = False) -> S3Object:
        r"""Download content from AWS S3 to local path.

        Args:
            force: Whether to force download even if local path already exists.

        Returns:
            self: This instance.

        """
        if self.key.endswith("/"):
            self.s3.download_files(
                bucket=self.bucket, prefix=self.key, path=self.path, overwrite=force
            )
        else:
            try:
                self.s3.download_file(
                    bucket=self.bucket, key=self.key, path=self.path, overwrite=force
                )
            except FileExistsError:
                pass
        return self

    def push(self, force: bool = False) -> S3Object:
        r"""Upload content of local path to AWS S3.

        Args:
            force: Whether to force upload even if S3 object already exists.

        Returns:
            self: This instance.

        """
        if self.key.endswith("/"):
            self.s3.upload_files(
                bucket=self.bucket, prefix=self.key, path=self.path, overwrite=force
            )
        else:
            try:
                self.s3.upload_file(
                    bucket=self.bucket, key=self.key, path=self.path, overwrite=force
                )
            except FileExistsError:
                pass
        return self

    def read_bytes(self) -> bytes:
        r"""Read file content from local path if it exists, or corresponding S3 object otherwise.

        Returns:
            Binary file content of storage object.

        """
        assert not self.key.endswith("/"), "S3 object key must referrence a file object"
        if self.path.is_file():
            return self.path.read_bytes()
        return self.s3.read_bytes(bucket=self.bucket, key=self.key)

    def write_bytes(self, data: bytes) -> S3Object:
        r"""Write bytes directly to storage object.

        Args:
            data: Binary data to write.

        Returns:
            self: This storage object.

        """
        assert not self.key.endswith("/"), "S3 object key must referrence a file object"
        self.s3.write_bytes(bucket=self.bucket, key=self.key, data=data)
        return self

    def read_text(self, encoding: Optional[str] = None) -> str:
        r"""Read text file content from local path if it exists, or corresponding S3 object otherwise.

        Args:
            encoding: Text encoding.

        Returns:
            text: Decoded text file content of storage object.

        """
        data = self.read_bytes()
        return data.decode(encoding) if encoding else data.decode()

    def write_text(self, text: str, encoding: Optional[str] = None) -> S3Object:
        r"""Write text to storage object.

        Args:
            text: Text to write.
            encoding: Text encoding.

        Returns:
            self: This storage object.

        """
        self.write_bytes(text.encode(encoding) if encoding else text.encode())
        return self

    def rmdir(self) -> S3Object:
        r"""Remove S3 objects both locally and from cloud storage."""
        if not self.key.endswith("/"):
            raise NotADirectoryError("S3 object key prefix must end with '/' character")
        super().rmdir()
        self.s3.delete_files(bucket=self.bucket, prefix=self.key)
        return self

    def unlink(self) -> S3Object:
        r"""Remove S3 object both locally and from cloud storage."""
        assert not self.key.endswith("/"), "S3 object key must not end with '/' character"
        super().unlink()
        self.s3.delete_file(bucket=self.bucket, key=self.key)
        return self

    def __repr__(self) -> str:
        r"""Get human-readable string representation of storage object reference."""
        return type(self).__name__ + "(bucket='{}', key='{}', path='{}')".format(
            self.bucket, self.key, self.path
        )



import pytest

import numpy as np
import torch
from torch import Tensor

from deepali.core import affine as A
from deepali.core.grid import Grid


@pytest.fixture
def default_angle() -> Tensor:
    return torch.deg2rad(torch.tensor(33.0))


@pytest.fixture
def default_grid(default_angle: Tensor) -> Grid:
    direction = A.euler_rotation_matrix(default_angle)
    grid = Grid(size=(34, 42), spacing=(0.7, 1.2), center=(7, 4), direction=direction)
    return grid


def test_grid_to_from_numpy(default_grid: Grid) -> None:
    r"""Test converting a Grid to a 1-dimensional NumPy array and constructing a new one from such array."""
    dim = default_grid.ndim
    arr = default_grid.numpy()
    assert isinstance(arr, np.ndarray)
    assert arr.ndim == 1
    assert arr.dtype == (np.float32 if default_grid.dtype == torch.float32 else np.float64)
    assert arr.shape == ((dim + 3) * dim,)
    grid = Grid.from_numpy(arr)
    assert grid == default_grid


def test_grid_eq(default_grid: Grid, default_angle: Tensor) -> None:
    r"""Test comparison of different Grid instances for equality."""
    # Same instance
    assert default_grid == default_grid

    # Different instance, same attributes
    other_grid = Grid(
        size=default_grid.size(),
        spacing=default_grid.spacing(),
        center=default_grid.center(),
        direction=default_grid.direction(),
    )
    assert default_grid == other_grid

    # Different size
    other_grid = Grid(
        size=default_grid.size_tensor().add(1),
        spacing=default_grid.spacing(),
        center=default_grid.center(),
        direction=default_grid.direction(),
    )
    assert default_grid != other_grid

    # Different spacing
    other_grid = Grid(
        size=default_grid.size(),
        spacing=default_grid.spacing().add(0.001),
        center=default_grid.center(),
        direction=default_grid.direction(),
    )
    assert default_grid != other_grid

    # Different center
    other_grid = Grid(
        size=default_grid.size(),
        spacing=default_grid.spacing(),
        center=default_grid.center().add(0.001),
        direction=default_grid.direction(),
    )
    assert default_grid != other_grid

    # Different direction
    other_direction = A.euler_rotation_matrix(default_angle.add(0.001))

    other_grid = Grid(
        size=default_grid.size(),
        spacing=default_grid.spacing(),
        center=default_grid.center(),
        direction=other_direction,
    )
    assert default_grid != other_grid


def test_grid_crop():
    grid = Grid((10, 7, 5))

    new_grid = grid.crop(2, -1)
    assert isinstance(new_grid, Grid)
    assert new_grid is not grid
    assert new_grid.size() == (6, 9, 5)
    assert torch.allclose(new_grid.center(), grid.center())
    assert torch.allclose(new_grid.index_to_world((-2, 1, 0)), grid.origin())

    new_grid_2 = grid.crop((2, -1))
    assert new_grid_2 == new_grid

    new_grid_2 = grid.crop(margin=(2, -1))
    assert new_grid_2 == new_grid

    new_grid_2 = grid.crop(num=(2, 2, -1, -1))
    assert new_grid_2 == new_grid


def test_grid_pad():
    grid = Grid((10, 7, 5))

    new_grid = grid.pad(4, -1)
    assert isinstance(new_grid, Grid)
    assert new_grid is not grid
    assert new_grid.size() == (18, 5, 5)
    assert torch.allclose(new_grid.center(), grid.center())
    assert torch.allclose(new_grid.index_to_world((4, -1, 0)), grid.origin())

    new_grid_2 = grid.pad((4, -1))
    assert new_grid_2 == new_grid

    new_grid_2 = grid.pad(margin=(4, -1))
    assert new_grid_2 == new_grid

    new_grid_2 = grid.pad(num=(4, 4, -1, -1))
    assert new_grid_2 == new_grid



r"""Linear transformation models."""

from __future__ import annotations

from collections import OrderedDict
import math
from typing import Callable, Optional, Union

import torch
from torch import Size, Tensor
from torch.nn import init

from deepali.core import affine as U
from deepali.core.grid import Grid
from deepali.core.linalg import normalize_quaternion
from deepali.core.linalg import quaternion_to_rotation_matrix
from deepali.core.linalg import rotation_matrix_to_quaternion
from deepali.core.tensor import as_float_tensor

from .base import LinearTransform
from .composite import SequentialTransform
from .parametric import InvertibleParametricTransform


class HomogeneousTransform(InvertibleParametricTransform, LinearTransform):
    r"""Arbitrary homogeneous coordinate transformation."""

    def __init__(
        self: HomogeneousTransform,
        grid: Grid,
        groups: Optional[int] = None,
        params: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Grid domain on which transformation is defined.
            groups: Number of transformations. Must be either 1 or equal to batch size.
            params: Homogeneous transform as tensor of shape ``(N, D, D + 1)``.

        """
        super().__init__(grid, groups=groups, params=params)

    @property
    def data_shape(self: HomogeneousTransform) -> Size:
        r"""Get shape of transformation parameters tensor."""
        return Size((self.ndim, self.ndim + 1))

    def matrix_(self: HomogeneousTransform, arg: Tensor) -> HomogeneousTransform:
        r"""Set transformation matrix."""
        if not isinstance(arg, Tensor):
            raise TypeError("HomogeneousTransform.matrix() 'arg' must be tensor")
        if arg.ndim != 3:
            raise ValueError("HomogeneousTransform.matrix() 'arg' must be 3-dimensional tensor")
        return self.data_(arg)

    def tensor(self: HomogeneousTransform) -> Tensor:
        r"""Get tensor representation of this transformation

        Returns:
            Batch of homogeneous transformation matrices as tensor of shape ``(N, D, D + 1)``.

        """
        matrix = self.data()
        if self.invert:
            N = matrix.shape[0]
            D = matrix.shape[1]
            row = torch.zeros((1, 1, D + 1), dtype=matrix.dtype, device=matrix.device)
            row[..., -1] = 1
            matrix = torch.cat((matrix, row.expand(N, 1, D + 1)), dim=1)
            matrix = torch.inverse(matrix)
            matrix = matrix.narrow(1, 0, D)
            matrix = matrix
        return matrix

    def extra_repr(self: HomogeneousTransform) -> str:
        r"""Print current transformation."""
        s = super().extra_repr() + ", matrix="
        if self.params is None:
            s += "undef"
        else:
            s += f"{self.matrix().tolist()!r}"
        return s


class Translation(InvertibleParametricTransform, LinearTransform):
    r"""Translation."""

    def __init__(
        self: Translation,
        grid: Grid,
        groups: Optional[int] = None,
        params: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Grid domain on which transformation is defined.
            groups: Number of transformations. Must be either 1 or equal to batch size.
            params: Translation offsets as tensor of shape ``(N, D)`` or ``(N, D, 1)``.

        """
        super().__init__(grid, groups=groups, params=params)

    @property
    def data_shape(self: Translation) -> Size:
        r"""Get shape of transformation parameters tensor."""
        return Size((self.ndim,))

    def offset(self: Translation) -> Tensor:
        r"""Get current translation offset in cube units."""
        return self.data()

    @torch.no_grad()
    def offset_(self: Translation, arg: Tensor) -> Translation:
        r"""Reset parameters to given translation in cube units."""
        if not isinstance(arg, Tensor):
            raise TypeError("Translation.offset() 'arg' must be tensor")
        params = as_float_tensor(arg)
        if params.isnan().any() or params.isinf().any():
            raise ValueError("Translation.offset() 'arg' must not be nan or inf")
        self.data_(params)
        return self

    def tensor(self: Translation) -> Tensor:
        r"""Get tensor representation of this transformation

        Returns:
            Batch of homogeneous transformation matrices as tensor of shape ``(N, D, 1)``.

        """
        offset = self.offset()
        if self.invert:
            offset = -offset
        return U.translation(offset)

    def extra_repr(self: Translation) -> str:
        r"""Print current transformation."""
        s = super().extra_repr() + ", offset="
        if self.params is None:
            s += "undef"
        else:
            s += f"{self.offset().tolist()!r}"
        return s


class EulerRotation(InvertibleParametricTransform, LinearTransform):
    r"""Euler rotation."""

    def __init__(
        self: EulerRotation,
        grid: Grid,
        groups: Optional[int] = None,
        params: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
        order: Optional[str] = None,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Grid domain on which transformation is defined.
            groups: Number of transformations. Must be 1 or equal to batch size.
            params: Rotation angles in degrees. This parameterization is adopted from MIRTK
                and ensures that rotation angles and scaling factors (percentage) are within
                a similar range of magnitude which is useful for direct optimization of these
                parameters. If parameters are predicted by a callable ``torch.nn.Module``,
                different output activations may be chosen before converting these to degrees.
            order: Order in which to compose elementary rotations. For example in 3D, "zxz" means
                that the first rotation occurs about z, the second about x, and the third rotation
                about z again. In 2D, this argument is ignored and a single rotation about z
                (plane normal) is applied.

        """
        if grid.ndim < 2 or grid.ndim > 3:
            raise ValueError("EulerRotation() 'grid' must be 2- or 3-dimensional")
        super().__init__(grid, groups=groups, params=params)
        self.order = order

    @property
    def data_shape(self: EulerRotation) -> Size:
        r"""Get shape of transformation parameters tensor."""
        return Size((self.nangles,))

    @property
    def nangles(self: EulerRotation) -> int:
        r"""Number of Euler angles."""
        return 1 if self.ndim == 2 else self.ndim

    def angles(self: EulerRotation) -> Tensor:
        r"""Get Euler angles in radians."""
        params = self.data()
        if self.has_parameters():
            params = params.tanh().mul(math.pi)
        return params

    @torch.no_grad()
    def angles_(self: EulerRotation, arg: Tensor) -> EulerRotation:
        r"""Reset parameters to given Euler angles in radians."""
        if not isinstance(arg, Tensor):
            raise TypeError("EulerRotation.angles() 'arg' must be tensor")
        shape = self.data_shape
        if arg.ndim != len(shape) + 1:
            raise ValueError(
                f"EulerRotation.angles() 'arg' must be {len(shape) + 1}-dimensional tensor"
            )
        shape = (arg.shape[0],) + self.data_shape
        if arg.shape != shape:
            raise ValueError(f"EulerRotation.angles() 'arg' must have shape {shape!r}")
        params = as_float_tensor(arg)
        if self.has_parameters():
            params = params.div(math.pi).atanh()
            if params.isnan().any():
                raise ValueError("EulerRotation.angles() 'arg' must be in range [-pi, pi]")
        self.data_(params)
        return self

    def matrix_(self: EulerRotation, arg: Tensor) -> EulerRotation:
        r"""Set rotation angles from rotation matrix."""
        if not isinstance(arg, Tensor):
            raise TypeError("EulerRotation.matrix() 'arg' must be tensor")
        if arg.ndim != 3:
            raise ValueError("EulerRotation.matrix() 'arg' must be 3-dimensional tensor")
        shape = (arg.shape[0], 3, 3)
        if arg.shape != shape:
            raise ValueError(f"Rotation matrix must have shape {shape!r}")
        angles = U.euler_rotation_angles(arg, order=self.order)
        return self.angles_(angles)

    def tensor(self: EulerRotation) -> Tensor:
        r"""Get tensor representation of this transformation

        Returns:
            Batch of homogeneous transformation matrices as tensor of shape ``(N, D, D)``.

        """
        mat = U.euler_rotation_matrix(self.angles(), order=self.order)
        if self.invert:
            mat = mat.transpose(1, 2)
        return mat

    def extra_repr(self: EulerRotation) -> str:
        r"""Print current transformation."""
        s = super().extra_repr() + ", angles="
        if self.params is None:
            s += "undef"
        else:
            s += f"{self.angles().tolist()!r}"
        return s


class QuaternionRotation(InvertibleParametricTransform, LinearTransform):
    r"""Quaternion based rotation in 3D."""

    def __init__(
        self: QuaternionRotation,
        grid: Grid,
        groups: Optional[int] = None,
        params: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Grid domain on which transformation is defined.
            groups: Number of transformations. Must be either 1 or equal to batch size.
            params: (nnormalized quaternion as 2-dimensional tensor of ``(N, 4)``.

        """
        if grid.ndim != 3:
            raise ValueError("QuaternionRotation() 'grid' must be 3-dimensional")
        super().__init__(grid, groups=groups, params=params)

    @property
    def data_shape(self: QuaternionRotation) -> Size:
        r"""Get shape of transformation parameters tensor."""
        return Size((4,))

    @torch.no_grad()
    def reset_parameters(self: QuaternionRotation) -> None:
        r"""Reset transformation parameters."""
        params = self.params
        if isinstance(params, Tensor):
            params.copy_(torch.tensor([0, 0, 0, 1], dtype=params.dtype, device=params.device))

    def quaternion(self: QuaternionRotation) -> Tensor:
        r"""Get rotation quaternion."""
        params = self.data()
        return normalize_quaternion(params)

    @torch.no_grad()
    def quaternion_(self: QuaternionRotation, arg: Tensor) -> QuaternionRotation:
        r"""Set rotation quaternion."""
        if not isinstance(arg, Tensor):
            raise TypeError("QuaternionRotation.quaternion() 'arg' must be tensor")
        shape = self.data_shape
        if arg.ndim != len(shape) + 1:
            raise ValueError(
                f"QuaternionRotation.quaternion() 'arg' must be {len(shape) + 1}-dimensional tensor"
            )
        shape = (arg.shape[0],) + self.data_shape
        if arg.shape != shape:
            raise ValueError(f"QuaternionRotation.quaternion() 'arg' must have shape {shape!r}")
        params = as_float_tensor(arg)
        params = normalize_quaternion(params)
        self.data_(params)
        return self

    def matrix_(self: QuaternionRotation, arg: Tensor) -> QuaternionRotation:
        r"""Set rotation quaternion from rotation matrix."""
        if not isinstance(arg, Tensor):
            raise TypeError("QuaternionRotation.matrix() 'arg' must be tensor")
        if arg.ndim != 3:
            raise ValueError("QuaternionRotation.matrix() 'arg' must be 3-dimensional tensor")
        shape = (arg.shape[0], 3, 3)
        if arg.shape != shape:
            raise ValueError(f"Rotation matrix must have shape {shape!r}")
        arg = rotation_matrix_to_quaternion(arg)
        return self.quaternion_(arg)

    def tensor(self: QuaternionRotation) -> Tensor:
        r"""Get tensor representation of this transformation

        Returns:
            Batch of homogeneous transformation matrices as tensor of shape ``(N, D, D)``.

        """
        q = self.data()
        mat = quaternion_to_rotation_matrix(q)
        if self.invert:
            mat = mat.transpose(1, 2)
        return mat

    def extra_repr(self: QuaternionRotation) -> str:
        r"""Print current transformation."""
        s = super().extra_repr() + ", q="
        if self.params is None:
            s += "undef"
        else:
            s += f"{self.quaternion().tolist()!r}"
        return s


class IsotropicScaling(InvertibleParametricTransform, LinearTransform):
    r"""Isotropic scaling."""

    def __init__(
        self: IsotropicScaling,
        grid: Grid,
        groups: Optional[int] = None,
        params: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Grid domain on which transformation is defined.
            groups: Number of transformations. Must be either 1 or equal to batch size.
            params: Isotropic scaling factor as a percentage. This parameterization is
                adopted from MIRTK and ensures that rotation angles in degrees and scaling
                factors are within a similar range of magnitude which is useful for direct
                optimization of these parameters. If parameters are predicted by a callable
                ``torch.nn.Module``, different output activations may be chosen before
                converting these to percentages (i.e., multiplied by 100).

        """
        super().__init__(grid, groups=groups, params=params)

    @property
    def data_shape(self: IsotropicScaling) -> Size:
        r"""Get shape of transformation parameters tensor."""
        return Size((1,))

    @torch.no_grad()
    def reset_parameters(self: IsotropicScaling) -> None:
        r"""Reset transformation parameters."""
        params = self.params
        if isinstance(params, Tensor):
            init.constant_(params, 1)

    def scales(self: IsotropicScaling) -> Tensor:
        r"""Get scaling factors."""
        params = self.data()
        if self.has_parameters():
            params = params.sub(1).tanh().exp()
        return params

    @torch.no_grad()
    def scales_(self: IsotropicScaling, arg: Tensor) -> IsotropicScaling:
        r"""Set transformation parameters from scaling factors."""
        if not isinstance(arg, Tensor):
            raise TypeError("IsotropicScaling.scales() 'arg' must be tensor")
        shape = self.data_shape
        if arg.ndim != len(shape) + 1:
            raise ValueError(
                f"IsotropicScaling.scales() 'arg' must be {len(shape) + 1}-dimensional tensor"
            )
        shape = (arg.shape[0],) + shape
        if arg.shape != shape:
            raise ValueError(f"IsotropicScaling.scales() 'arg' must have shape {shape!r}")
        params = as_float_tensor(arg)
        if self.has_parameters():
            params = params.log().atanh().add(1)
            if params.isnan().any():
                raise ValueError("IsotropicScaling.scales() 'arg' must be positive")
        self.data_(params)
        return self

    def tensor(self: IsotropicScaling) -> Tensor:
        r"""Get tensor representation of this transformation

        Returns:
            Batch of homogeneous transformation matrices as tensor of shape ``(N, D, D)``.

        """
        scales = self.scales()
        if self.invert:
            scales = 1 / scales
        return U.scaling_transform(scales.expand(scales.shape[0], self.ndim))

    def extra_repr(self: IsotropicScaling) -> str:
        """Print current transformation."""
        s = super().extra_repr() + ", scales="
        if self.params is None:
            s += "undef"
        else:
            s += f"{self.scales().tolist()!r}"
        return s


class AnisotropicScaling(InvertibleParametricTransform, LinearTransform):
    r"""Anisotropic scaling."""

    def __init__(
        self: AnisotropicScaling,
        grid: Grid,
        groups: Optional[int] = None,
        params: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Grid domain on which transformation is defined.
            groups: Number of transformations. Must be either 1 or equal to batch size.
            params: Anisotropic scaling factors as percentages. This parameterization is
                adopted from MIRTK and ensures that rotation angles in degrees and scaling
                factors are within a similar range of magnitude which is useful for direct
                optimization of these parameters. If parameters are predicted by a callable
                ``torch.nn.Module``, different output activations may be chosen before
                converting these to percentages (i.e., multiplied by 100).

        """
        super().__init__(grid, groups=groups, params=params)

    @property
    def data_shape(self: AnisotropicScaling) -> Size:
        r"""Get shape of transformation parameters tensor."""
        return Size((self.ndim,))

    @torch.no_grad()
    def reset_parameters(self: AnisotropicScaling) -> None:
        r"""Reset transformation parameters."""
        params = self.params
        if isinstance(params, Tensor):
            init.constant_(params, 1)

    def scales(self: AnisotropicScaling) -> Tensor:
        r"""Get scaling factors."""
        params = self.data()
        if self.has_parameters():
            params = params.sub(1).tanh().exp()
        return params

    @torch.no_grad()
    def scales_(self: AnisotropicScaling, arg: Tensor) -> AnisotropicScaling:
        r"""Set transformation parameters from scaling factors."""
        if not isinstance(arg, Tensor):
            raise TypeError("AnisotropicScaling.scales() 'arg' must be tensor")
        shape = self.data_shape
        if arg.ndim != len(shape) + 1:
            raise ValueError(
                f"AnisotropicScaling.scales() 'arg' must be {len(shape) + 1}-dimensional tensor"
            )
        shape = (arg.shape[0],) + shape
        if arg.shape != shape:
            raise ValueError(f"AnisotropicScaling.scales() 'arg' must have shape {shape!r}")
        params = as_float_tensor(arg)
        if self.has_parameters():
            params = params.log().atanh().add(1)
            if params.isnan().any():
                raise ValueError("AnisotropicScaling.scales() 'arg' must be positive")
        self.data_(params)
        return self

    def tensor(self: AnisotropicScaling) -> Tensor:
        r"""Get tensor representation of this transformation

        Returns:
            Batch of homogeneous transformation matrices as tensor of shape ``(N, D, D)``.

        """
        scales = self.scales()
        if self.invert:
            scales = 1 / scales
        return U.scaling_transform(scales)

    def extra_repr(self: AnisotropicScaling) -> str:
        r"""Print current transformation."""
        s = super().extra_repr() + ", scales="
        if self.params is None:
            s += "undef"
        else:
            s += f"{self.scales().tolist()!r}"
        return s


class Shearing(InvertibleParametricTransform, LinearTransform):
    r"""Shear transformation."""

    def __init__(
        self: Shearing,
        grid: Grid,
        groups: Optional[int] = None,
        params: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Grid domain on which transformation is defined.
            groups: Number of transformations. Must be either 1 or equal to batch size.
            params: Shearing angles in degrees. This parameterization is adopted from MIRTK
                and ensures that rotation angles and scaling factors (percentage) are within
                a similar range of magnitude which is useful for direct optimization of these
                parameters. If parameters are predicted by a callable ``torch.nn.Module``,
                different output activations may be chosen before converting these to degrees.

        """
        if grid.ndim < 2 or grid.ndim > 3:
            raise ValueError("Shearing() 'grid' must be 2- or 3-dimensional'")
        super().__init__(grid, groups=groups, params=params)

    @property
    def data_shape(self: Shearing) -> Size:
        r"""Get shape of transformation parameters tensor."""
        return Size((self.nangles,))

    @property
    def nangles(self: Shearing) -> int:
        r"""Number of shear angles."""
        return 1 if self.ndim == 2 else self.ndim

    def angles(self: Shearing) -> Tensor:
        r"""Get shear angles in radians."""
        params = self.data()
        if self.has_parameters():
            params = params.tanh().mul(math.pi / 4)
        return params

    @torch.no_grad()
    def angles_(self: Shearing, arg: Tensor) -> Shearing:
        r"""Set transformation parameters from shear angles in radians."""
        if not isinstance(arg, Tensor):
            raise TypeError("Shearing.angles_() 'arg' must be tensor")
        shape = self.data_shape
        if arg.ndim != len(shape) + 1:
            raise ValueError(
                f"Shearing.angles_() 'arg' must be {len(shape) + 1}-dimensional tensor"
            )
        shape = (arg.shape[0],) + shape
        if arg.shape != shape:
            raise ValueError(f"Shearing.angles_() 'arg' must have shape {shape!r}")
        params = as_float_tensor(arg)
        if self.has_parameters():
            params = params.mul(4 / math.pi).atanh()
            if params.isnan().any():
                raise ValueError("Shear 'angles' must be in range [-pi/4, pi/4]")
        self.data_(params)
        return self

    def tensor(self: Shearing) -> Tensor:
        r"""Get tensor representation of this transformation

        Returns:
            Batch of homogeneous transformation matrices as tensor of shape ``(N, D, D)``.

        """
        mat = U.shear_matrix(self.angles())
        if self.invert:
            mat = torch.inverse(mat)
        return mat

    def extra_repr(self: Shearing) -> str:
        r"""Print current transformation."""
        s = super().extra_repr() + "angles="
        if self.params is None:
            s += "undef"
        else:
            s += f"{self.angles().tolist()!r}"
        return s


class RigidTransform(SequentialTransform):
    r"""Rigid transformation."""

    def __init__(
        self: RigidTransform,
        grid: Grid,
        groups: Optional[int] = None,
        rotation: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
        translation: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Domain with respect to which transformation is defined.
            groups: Number of transformations ``N``.
            rotation: Parameters of ``EulerRotation``.
            translation: Parameters of ``Translation``.

        """
        transforms = OrderedDict()
        transforms["rotation"] = EulerRotation(grid, groups=groups, params=rotation)
        transforms["translation"] = Translation(grid, groups=groups, params=translation)
        super().__init__(transforms)

    @property
    def rotation(self) -> EulerRotation:
        return self._transforms["rotation"]

    @property
    def translation(self) -> Translation:
        return self._transforms["translation"]


class RigidQuaternionTransform(SequentialTransform):
    r"""Rigid transformation parameterized by rotation quaternion."""

    def __init__(
        self: RigidQuaternionTransform,
        grid: Grid,
        groups: Optional[int] = None,
        rotation: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
        translation: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Domain with respect to which transformation is defined.
            groups: Number of transformations ``N``.
            rotation: Parameters of ``QuaternionRotation``.
            translation: Parameters of ``Translation``.

        """
        transforms = OrderedDict()
        transforms["rotation"] = QuaternionRotation(grid, groups=groups, params=rotation)
        transforms["translation"] = Translation(grid, groups=groups, params=translation)
        super().__init__(transforms)

    @property
    def rotation(self) -> QuaternionRotation:
        return self._transforms["rotation"]

    @property
    def translation(self) -> Translation:
        return self._transforms["translation"]


class SimilarityTransform(SequentialTransform):
    r"""Similarity transformation with isotropic scaling."""

    def __init__(
        self: SimilarityTransform,
        grid: Grid,
        groups: Optional[int] = None,
        scaling: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
        rotation: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
        translation: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Domain with respect to which transformation is defined.
            groups: Number of transformations ``N``.
            scaling: Parameters of ``IsotropicScaling``.
            rotation: Parameters of ``EulerRotation``.
            translation: Parameters of ``Translation``.

        """
        transforms = OrderedDict()
        transforms["scaling"] = IsotropicScaling(grid, groups=groups, params=scaling)
        transforms["rotation"] = EulerRotation(grid, groups=groups, params=rotation)
        transforms["translation"] = Translation(grid, groups=groups, params=translation)
        super().__init__(transforms)

    @property
    def rotation(self) -> EulerRotation:
        return self._transforms["rotation"]

    @property
    def scaling(self) -> IsotropicScaling:
        return self._transforms["scaling"]

    @property
    def translation(self) -> Translation:
        return self._transforms["translation"]


class AffineTransform(SequentialTransform):
    r"""Affine transformation without shearing."""

    def __init__(
        self: AffineTransform,
        grid: Grid,
        groups: Optional[int] = None,
        scaling: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
        rotation: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
        translation: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Domain with respect to which transformation is defined.
            groups: Number of transformations ``N``.
            scaling: Parameters of ``AnisotropicScaling``.
            rotation: Parameters of ``EulerRotation``.
            translation: Parameters of ``Translation``.

        """
        transforms = OrderedDict()
        transforms["scaling"] = AnisotropicScaling(grid, groups=groups, params=scaling)
        transforms["rotation"] = EulerRotation(grid, groups=groups, params=rotation)
        transforms["translation"] = Translation(grid, groups=groups, params=translation)
        super().__init__(transforms)

    @property
    def rotation(self) -> EulerRotation:
        return self._transforms["rotation"]

    @property
    def scaling(self) -> AnisotropicScaling:
        return self._transforms["scaling"]

    @property
    def translation(self) -> Translation:
        return self._transforms["translation"]


class FullAffineTransform(SequentialTransform):
    r"""Affine transformation including shearing."""

    def __init__(
        self: FullAffineTransform,
        grid: Grid,
        groups: Optional[int] = None,
        scaling: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
        shearing: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
        rotation: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
        translation: Optional[Union[bool, Tensor, Callable[..., Tensor]]] = True,
    ) -> None:
        r"""Initialize transformation parameters.

        Args:
            grid: Domain with respect to which transformation is defined.
            groups: Number of transformations ``N``.
            scaling: Parameters of ``AnisotropicScaling``.
            shearing: Parameters of ``Shearing``.
            rotation: Parameters of ``EulerRotation``.
            translation: Parameters of ``Translation``.

        """
        transforms = OrderedDict()
        transforms["scaling"] = AnisotropicScaling(grid, groups=groups, params=scaling)
        transforms["shearing"] = Shearing(grid, groups=groups, params=shearing)
        transforms["rotation"] = EulerRotation(grid, groups=groups, params=rotation)
        transforms["translation"] = Translation(grid, groups=groups, params=translation)
        super().__init__(transforms)

    @property
    def rotation(self) -> EulerRotation:
        return self._transforms["rotation"]

    @property
    def scaling(self) -> AnisotropicScaling:
        return self._transforms["scaling"]

    @property
    def shearing(self) -> Shearing:
        return self._transforms["shearing"]

    @property
    def translation(self) -> Translation:
        return self._transforms["translation"]



r"""Auxiliary functions for reading and writing NIfTI image files."""

import sys
from typing import Tuple

try:
    import nibabel as nib
except ImportError:
    nib = None

import numpy as np
import torch
from torch import Tensor

from deepali.core.grid import Grid
from deepali.core.pathlib import PathUri, path_suffix, unlink_or_mkdir
from deepali.core.storage import StorageObject


NIFTI_IMAGE_SUFFIXES = (".nia", ".nii", ".nii.gz", ".hdr", ".hdr.gz", ".img", ".img.gz")


def has_nifti_image_suffix(path: PathUri) -> bool:
    r"""Check whether filename ends in a NIfTI image extension."""
    return path_suffix(path).lower() in NIFTI_IMAGE_SUFFIXES


def read_nifti_image(path: PathUri) -> Tuple[Tensor, Grid]:
    r"""Read image data and grid attributes from NIfTI file."""
    if not has_nifti_image_suffix(path):
        raise ValueError(f"NIfTI image filename suffix must be one of {NIFTI_IMAGE_SUFFIXES}")
    if nib is None:
        try:
            from .sitk import read_sitk_image
        except ImportError:
            raise RuntimeError("nibabel or SimpleITK is required for reading NIfTI images")
        return read_sitk_image(path)
    # Load image
    with StorageObject.from_path(path) as obj:
        obj.pull(force=True)
        image = nib.load(obj.path)
    # Image sampling grid attributes
    dim = np.asarray(image.header["dim"])
    ndim = int(dim[0])
    D = min(ndim, 3)
    size = dim[1 : D + 1]
    spacing = np.asarray(image.header["pixdim"][1 : D + 1])
    affine = np.asarray(image.affine)
    origin = affine[:D, 3]
    direction = np.divide(affine[:D, :D], spacing)
    # Convert to ITK LPS convention
    origin[:2] *= -1
    direction[:2] *= -1
    # Replace small values and -0 by 0
    epsilon = sys.float_info.epsilon
    origin[np.abs(origin) < epsilon] = 0
    direction[np.abs(direction) < epsilon] = 0
    # Image data array
    slope = image.dataobj.slope
    inter = image.dataobj.inter
    if abs(slope) > epsilon and (abs(slope - 1) > epsilon or abs(inter) > epsilon):
        data: np.ndarray = image.get_fdata()
    else:
        data: np.ndarray = image.dataobj.get_unscaled()
    # Squeeze unused dimensions
    # https://github.com/InsightSoftwareConsortium/ITK/blob/3454d857dc46e4333ad1178be8c186547fba87ef/Modules/IO/NIFTI/src/itkNiftiImageIO.cxx#L1112-L1156
    intent_code = int(image.header["intent_code"])
    if intent_code in (1005, 1006, 1007):
        # Vector or matrix valued image
        for realdim in range(4, 1, -1):
            if dim[realdim] > 1:
                break
        else:
            realdim = 1
    elif intent_code == 1004:
        raise NotImplementedError(
            f"{path} has an intent code of NIFTI_INTENT_GENMATRIX which is not yet implemented"
        )
    else:
        # Scalar image
        realdim = ndim
        while realdim > 3 and dim[realdim] == 1:
            realdim -= 1
    data = np.reshape(data, data.shape[:realdim] + data.shape[5:])
    # Reverse order of axes
    data = np.transpose(data, axes=tuple(reversed(range(data.ndim))))
    # Add leading channel dimension
    grid = Grid(size=size, origin=origin, spacing=spacing, direction=direction)
    if data.ndim == grid.ndim:
        data = np.expand_dims(data, 0)
    if data.dtype == np.uint16:
        data = data.astype(np.int32)
    elif data.dtype == np.uint32:
        data = data.astype(np.int64)
    return torch.from_numpy(data), grid


def write_nifti_image(data: Tensor, grid: Grid, path: PathUri) -> None:
    r"""Write image data and grid attributes to NIfTI image file."""
    if not has_nifti_image_suffix(path):
        raise ValueError(f"NIfTI image filename suffix must be one of {NIFTI_IMAGE_SUFFIXES}")
    if nib is None:
        try:
            from .sitk import write_sitk_image
        except ImportError:
            raise RuntimeError("nibabel or SimpleITK is required for writing NIfTI images")
        return write_sitk_image(data, grid, path, compress=str(path).lower().endswith(".gz"))
    data = data.detach().cpu()
    if data.ndim == grid.ndim:
        data = data.unsqueeze(0)
    if data.ndim != grid.ndim + 1:
        raise ValueError("write_image() data.ndim must be equal to grid.ndim or grid.ndim + 1")
    # Reverse order of axes
    dataobj = np.transpose(data.numpy(), axes=tuple(reversed(range(data.ndim))))
    # Convert to NIfTI RAS convention
    affine = grid.affine().cpu().numpy()
    affine[:2] *= -1
    with StorageObject.from_path(path) as obj:
        local_path = unlink_or_mkdir(obj.path)
        nib.save(nib.Nifti1Image(dataobj, affine), str(local_path))
        obj.push(force=True)



from copy import deepcopy
from typing import Tuple

import numpy as np
import pytest
import torch
from torch import Tensor

from deepali.core import Axes, Grid
from deepali.data import FlowField, FlowFields, Image, ImageBatch


def image_size(sdim: int) -> Tuple[int, ...]:
    if sdim == 2:
        return (64, 57)
    if sdim == 3:
        return (64, 57, 31)
    raise ValueError("image_size() 'sdim' must be 2 or 3")


def image_shape(sdim: int) -> Tuple[int, ...]:
    return tuple(reversed(image_size(sdim)))


@pytest.fixture(scope="function")
def grid(request) -> Tensor:
    size = image_size(request.param)
    spacing = (0.25, 0.2, 0.5)[: len(size)]
    return Grid(size=size, spacing=spacing)


@pytest.fixture(scope="function")
def data(request) -> Tensor:
    shape = image_shape(request.param)
    return torch.randn((len(shape),) + shape).mul_(100)


@pytest.fixture(scope="function")
def zeros(request) -> Tensor:
    shape = image_shape(request.param)
    return torch.zeros((len(shape),) + shape)


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (3,)], indirect=True)
def test_flowfield_torch_function(zeros: Tensor, grid: Grid) -> None:
    data = zeros
    axes = Axes.WORLD  # different from default to check if attribute is preserved

    image = FlowField(data, grid, axes)
    assert type(image) is FlowField
    assert hasattr(image, "_axes")
    assert image.axes() is axes
    assert hasattr(image, "_grid")
    assert image.grid() is grid

    result = image.type(image.dtype)
    assert result is image

    result = image.type(torch.int16)
    assert result is not image
    assert type(result) is FlowField
    assert hasattr(result, "_axes")
    assert result.axes() is axes
    assert hasattr(result, "_grid")
    assert result.grid() is image.grid()
    assert result.dtype is torch.int16
    assert result.data_ptr() != image.data_ptr()

    result = image.eq(0)
    assert type(result) is FlowField
    assert hasattr(result, "_axes")
    assert result.axes() is axes
    assert hasattr(result, "_grid")
    assert result.grid() is image.grid()
    assert result.dtype is torch.bool
    assert result.data_ptr() != image.data_ptr()

    result = result.all()
    assert type(result) is Tensor
    assert not hasattr(result, "_axes")
    assert not hasattr(result, "_grid")
    assert result.ndim == 0
    assert result.item() is True

    result = torch.add(image, 2)
    assert type(result) is FlowField
    assert hasattr(result, "_axes")
    assert result.axes() is image.axes()
    assert hasattr(result, "_grid")
    assert result.grid() is image.grid()
    assert result.eq(2).all()

    result = torch.add(4, image)
    assert type(result) is FlowField
    assert hasattr(result, "_axes")
    assert result.axes() is image.axes()
    assert hasattr(result, "_grid")
    assert result.grid() is image.grid()
    assert result.eq(4).all()

    result = image.add(1)
    assert type(result) is FlowField
    assert hasattr(result, "_axes")
    assert result.axes() is image.axes()
    assert hasattr(result, "_grid")
    assert result.grid() is image.grid()
    assert result.eq(1).all()

    result = image.clone()
    assert type(result) is FlowField
    assert hasattr(result, "_axes")
    assert result.axes() is image.axes()
    assert hasattr(result, "_grid")
    assert result.grid() is not image.grid()
    assert result.grid() == image.grid()
    assert result.data_ptr() != image.data_ptr()

    result = image.to("cpu", torch.int16)
    assert type(result) is FlowField
    assert hasattr(result, "_axes")
    assert result.axes() is image.axes()
    assert hasattr(result, "_grid")
    assert result.grid() is image.grid()
    assert result.device == torch.device("cpu")
    assert result.dtype == torch.int16

    if torch.cuda.is_available():
        result = image.cuda()
        assert type(result) is FlowField
        assert hasattr(result, "_axes")
        assert result.axes() is image.axes()
        assert hasattr(result, "_grid")
        assert result.grid() is image.grid()
        assert result.is_cuda

    result = torch.cat([image, image], dim=0)
    assert isinstance(result, Image)
    assert type(result) == Image
    assert not hasattr(result, "_axes")
    assert hasattr(result, "_grid")
    assert result.shape[0] == image.shape[0] * 2
    assert result.shape[1:] == image.shape[1:]
    assert result.grid() is image.grid()


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (3,)], indirect=True)
def test_flowfields_torch_function(zeros: Tensor, grid: Grid) -> None:
    data = zeros

    batch = FlowFields(data.unsqueeze(0), grid, Axes.WORLD)

    result = batch.detach()
    assert isinstance(result, FlowFields)
    assert result.shape == batch.shape
    assert result.axes() == batch.axes()
    assert result.grids() == batch.grids()

    result = torch.cat([batch, batch], dim=0)
    assert isinstance(result, FlowFields)
    assert result.shape[0] == batch.shape[0] + batch.shape[0]
    assert result.shape[1:] == batch.shape[1:]
    assert result.axes() == batch.axes()
    assert result.grids() == batch.grids() * 2

    result = torch.cat([batch, batch], dim=1)
    assert type(result) == ImageBatch
    assert result.shape[0] == batch.shape[0]
    assert result.shape[1] == batch.shape[1] * 2
    assert result.shape[2:] == batch.shape[2:]
    assert result.grids() == batch.grids()

    with pytest.raises(ValueError):
        # Cannot batch together flow fields with mismatching Axes
        torch.cat([batch, batch.axes(Axes.CUBE_CORNERS)], dim=0)


@pytest.mark.parametrize("zeros,grid", [(d, d) for d in (3,)], indirect=True)
def test_flowfields_getitem(zeros: Tensor, grid: Grid) -> None:
    grids = [deepcopy(grid) for _ in range(5)]  # make grids distinguishable
    batch = FlowFields(zeros.unsqueeze(0).expand((5,) + zeros.shape), grids, axes=Axes.GRID)
    for i in range(len(batch)):
        batch.tensor()[i] = i

    for i in range(len(batch)):
        item = batch[i]
        assert type(item) is FlowField
        assert item.axes() is batch.axes()
        assert item.grid() is batch.grid(i)
        assert torch.allclose(item, batch.tensor()[i])

    item = batch[1:-1:2]
    assert type(item) is FlowFields
    assert item.axes() is batch.axes()
    assert torch.allclose(item.tensor(), batch.tensor()[1:-1:2])

    item = batch[...]
    assert type(item) is FlowFields
    assert item.axes() is batch.axes()
    assert torch.allclose(item.tensor(), batch.tensor()[...])

    item = batch[:, ...]
    assert type(item) is FlowFields
    assert item.axes() is batch.axes()
    assert torch.allclose(item.tensor(), batch.tensor()[:, ...])

    item = batch[[4]]
    assert type(item) is FlowFields
    assert len(item) == 1
    assert item.axes() is batch.axes()
    assert item.grid(0) is batch.grid(4)
    assert torch.allclose(item.tensor(), batch.tensor()[[0]])

    index = [3, 2]
    item = batch[index]
    assert type(item) is FlowFields
    assert len(item) == 2
    assert item.axes() is batch.axes()
    assert item.grid(0) is batch.grid(3)
    assert item.grid(1) is batch.grid(2)
    assert torch.allclose(item.tensor(), batch.tensor()[index])

    index = (0, 2)
    item = batch[index]
    assert type(item) is Tensor
    assert torch.allclose(item, batch.tensor()[index])

    index = np.array([1, 3])
    item = batch[index]
    assert type(item) is FlowFields
    assert len(item) == 2
    assert item.axes() is batch.axes()
    assert item.grid(0) is batch.grid(1)
    assert item.grid(1) is batch.grid(3)
    assert torch.allclose(item.tensor(), batch.tensor()[index])

    index = torch.tensor([0, 2])
    item = batch[index]
    assert type(item) is FlowFields
    assert len(item) == 2
    assert item.axes() is batch.axes()
    assert item.grid(0) is batch.grid(0)
    assert item.grid(1) is batch.grid(2)
    assert torch.allclose(item.tensor(), batch.tensor()[index])

    item = batch[:, 0]
    assert type(item) is Tensor
    assert torch.allclose(item, batch.tensor()[:, 0])

    item = batch[:, 1:2]
    assert type(item) is ImageBatch
    assert all(a is b for a, b in zip(item.grids(), batch.grids()))
    assert torch.allclose(item, batch.tensor()[:, 1:2])

    if batch.ndim == 5:
        item = batch[:, :2, 0]
        assert type(item) is Tensor
        assert item.shape[1] == item.ndim - 2
        assert torch.allclose(item, batch.tensor()[:, :2, 0])

    item = batch[3, 1:2]
    assert type(item) is Image
    assert item.grid() is batch.grid(3)
    assert torch.allclose(item, batch.tensor()[3, 1:2])



from typing import Tuple

import pytest

import torch
from torch import Tensor

from deepali.core.typing import is_namedtuple
from deepali.networks.layers import convolution
from deepali.networks.unet import UNet, UNetConfig, last_num_channels


@pytest.fixture(scope="function")
def input_tensor(request) -> Tensor:
    if request.param == 2:
        return torch.rand((1, 1, 128, 128))
    elif request.param == 3:
        return torch.rand((1, 1, 64, 128, 128))
    else:
        raise ValueError("input_tensor() 'request.param' must be 2 or 3")


@pytest.mark.parametrize("input_tensor", [2, 3], indirect=True)
def test_unet_without_output_layer(input_tensor: Tensor) -> None:
    spatial_dims = input_tensor.ndim - 2
    in_channels = input_tensor.shape[1]

    model = UNet(spatial_dims=spatial_dims, in_channels=in_channels)
    model.eval()
    print(model)

    assert model.output_is_tensor() is False
    assert model.output_is_dict() is False
    assert model.output_is_tuple() is True

    output: Tuple[Tensor, ...] = model(input_tensor)
    assert isinstance(output, tuple)
    assert len(output) == 4
    assert isinstance(output[0], Tensor)
    assert isinstance(output[1], Tensor)
    assert isinstance(output[2], Tensor)
    assert isinstance(output[3], Tensor)

    for i, output_tensor in enumerate(output):
        ds = 2 ** (model.num_levels - 1 - i)
        channels = last_num_channels(model.num_channels[i])
        shape = (input_tensor.shape[0], channels) + tuple(n // ds for n in input_tensor.shape[2:])
        assert output_tensor.shape == shape


@pytest.mark.parametrize("input_tensor", [2, 3], indirect=True)
def test_unet_with_single_output_layer(input_tensor: Tensor) -> None:
    spatial_dims = input_tensor.ndim - 2
    in_channels = input_tensor.shape[1]
    out_channels = in_channels

    model = UNet(spatial_dims=spatial_dims, in_channels=in_channels, out_channels=out_channels)
    print(model)

    assert model.output_is_tensor() is True
    assert model.output_is_dict() is False
    assert model.output_is_tuple() is False

    output_tensor = model(input_tensor)
    assert isinstance(output_tensor, Tensor)
    assert output_tensor.shape[0] == input_tensor.shape[0]
    assert output_tensor.shape[1] == out_channels
    assert output_tensor.shape[2:] == input_tensor.shape[2:]


@pytest.mark.parametrize("input_tensor", [2, 3], indirect=True)
def test_unet_with_multiple_output_layers(input_tensor: Tensor) -> None:
    spatial_dims = input_tensor.ndim - 2
    config = UNetConfig()

    output_modules = {}
    output_indices = {}
    for i, channels in enumerate(config.decoder.num_channels):
        name = f"output_{i + 1}"
        output = convolution(
            spatial_dims=spatial_dims,
            in_channels=channels,
            out_channels=1,
            kernel_size=1,
            bias=False,
        )
        output_modules[name] = output
        output_indices[name] = i

    model = UNet(
        spatial_dims=spatial_dims,
        in_channels=1,
        output_modules=output_modules,
        output_indices=output_indices,
    )
    model.eval()
    print(model)

    assert model.output_is_tensor() is False
    assert model.output_is_dict() is True
    assert model.output_is_tuple() is False

    output = model(input_tensor)
    assert is_namedtuple(output)
    assert len(output) == len(output_modules)
    assert set(getattr(output, "_fields", [])) == set(output_modules.keys())
    assert set(output.keys()) == set(output_modules.keys())
    assert output["output_1"] is output.output_1
    assert all(isinstance(x, Tensor) for x in output)



r"""Engine for iterative optimization-based registration."""

# TODO: Use ignite.engine instead of custom RegistrationEngine.

from __future__ import annotations

import math
from collections import OrderedDict
from timeit import default_timer as timer
from typing import Callable

import torch
from torch import Tensor
from torch.optim import Optimizer
from torch.utils.hooks import RemovableHandle

from .optim import slope_of_least_squares_fit
from .losses import RegistrationLoss, RegistrationResult


PROFILING = False


class RegistrationEngine(object):
    r"""Minimize registration loss until convergence."""

    def __init__(
        self,
        loss: RegistrationLoss,
        optimizer: Optimizer,
        max_steps: int = 500,
        min_delta: float = 1e-6,
        min_value: float = float("nan"),
        max_history: int = 10,
    ):
        r"""Initialize registration loop."""
        self.loss = loss
        self.optimizer = optimizer
        self.num_steps = 0
        self.max_steps = max_steps
        self.min_delta = min_delta
        self.min_value = min_value
        self.max_history = max(2, max_history)
        self.loss_values = []
        self._eval_hooks = OrderedDict()
        self._step_hooks = OrderedDict()

    @property
    def loss_value(self) -> float:
        if not self.loss_values:
            return float("inf")
        return self.loss_values[-1]

    def step(self) -> float:
        r"""Perform one registration step.

        Returns:
            Loss value prior to taking gradient step.

        """
        num_evals = 0

        def closure() -> float:
            self.optimizer.zero_grad()
            t_start = timer()
            result = self.loss.eval()
            if PROFILING:
                print(f"Forward pass in {timer() - t_start:.3f}s")
            loss = result["loss"]
            assert isinstance(loss, Tensor)
            t_start = timer()
            loss.backward()
            if PROFILING:
                print(f"Backward pass in {timer() - t_start:.3f}s")
            nonlocal num_evals
            num_evals += 1
            with torch.no_grad():
                for hook in self._eval_hooks.values():
                    hook(self, self.num_steps, num_evals, result)
            return float(loss)

        loss_value = self.optimizer.step(closure)
        assert loss_value is not None

        with torch.no_grad():
            for hook in self._step_hooks.values():
                hook(self, self.num_steps, num_evals, loss_value)

        return loss_value

    def run(self) -> float:
        r"""Perform registration steps until convergence.

        Returns:
            Loss value prior to taking last gradient step.

        """
        self.loss_values = []
        self.num_steps = 0
        while self.num_steps < self.max_steps and not self.converged():
            value = self.step()
            self.num_steps += 1
            if math.isnan(value):
                raise RuntimeError(
                    f"NaN value in registration loss at gradient step {self.num_steps}"
                )
            if math.isinf(value):
                raise RuntimeError(
                    f"Inf value in registration loss at gradient step {self.num_steps}"
                )
            self.loss_values.append(value)
            if len(self.loss_values) > self.max_history:
                self.loss_values.pop(0)
        return self.loss_value

    def converged(self) -> bool:
        r"""Check convergence criteria."""
        values = self.loss_values
        if not values:
            return False
        value = values[-1]
        if self.min_delta < 0:
            epsilon = abs(self.min_delta * value)
        else:
            epsilon = self.min_delta
        slope = slope_of_least_squares_fit(values)
        if abs(slope) < epsilon:
            return True
        if value < self.min_value:
            return True
        return False

    def register_eval_hook(
        self, hook: Callable[[RegistrationEngine, int, int, RegistrationResult], None]
    ) -> RemovableHandle:
        r"""Registers a evaluation hook.

        The hook will be called every time after the registration loss has been evaluated
        during a single step of the optimizer, and the backward pass was performed, but
        before the model parameters are updated by taking a gradient step.

            hook(self, num_steps: int, num_evals: int, result: RegistrationResult) -> None

        Returns:
            A handle that can be used to remove the added hook by calling ``handle.remove()``

        """
        handle = RemovableHandle(self._eval_hooks)
        self._eval_hooks[handle.id] = hook
        return handle

    def register_step_hook(
        self, hook: Callable[[RegistrationEngine, int, int, float], None]
    ) -> RemovableHandle:
        r"""Registers a gradient step hook.

        The hook will be called every time after a gradient step of the optimizer.

            hook(self, num_steps: int, num_evals: int, loss: float) -> None

        Returns:
            A handle that can be used to remove the added hook by calling ``handle.remove()``

        """
        handle = RemovableHandle(self._step_hooks)
        self._step_hooks[handle.id] = hook
        return handle



r"""Image and spatial transformer network."""

from dataclasses import dataclass
from typing import Dict, Optional, Sequence, Union

import torch
from torch import Size, Tensor
from torch.nn import Module

from deepali.core import DataclassConfig, Grid, PaddingMode, Sampling
from deepali.core import functional as U
from deepali.spatial import GenericSpatialTransform, ImageTransformer, ParametricTransform

from .itn import ImageTransformerConfig, ImageTransformerNetwork
from .stn import SpatialTransformerConfig, SpatialTransformerNetwork


@dataclass
class InputConfig(DataclassConfig):
    size: Sequence[int]
    channels: int = 1

    @property
    def spatial_dims(self) -> int:
        return len(self.size)

    @property
    def shape(self) -> Size:
        return Size(reversed(self.size))


@dataclass
class ImageAndSpatialTransformerConfig(DataclassConfig):
    input: InputConfig
    itn: Optional[Union[str, ImageTransformerConfig]] = "miccai2019"
    stn: SpatialTransformerConfig = SpatialTransformerConfig()


class ImageAndSpatialTransformerNetwork(Module):
    r"""Image and spatial transformer network."""

    def __init__(
        self,
        stn: SpatialTransformerNetwork,
        itn: Optional[ImageTransformerNetwork] = None,
        sampling: Union[Sampling, str] = Sampling.LINEAR,
        padding: Union[PaddingMode, str, float] = PaddingMode.BORDER,
    ) -> None:
        super().__init__()
        if stn.in_channels < 2 or stn.in_channels % 2 != 0:
            raise ValueError(
                f"{type(self).__name__}() 'stn.in_channels' must be positive even number"
            )
        self.itn = itn
        self.stn = stn
        grid = Grid(size=stn.in_size)
        transform = GenericSpatialTransform(grid, params=None, config=stn.config)
        self.warp = ImageTransformer(transform, sampling=sampling, padding=padding)

    @property
    def config(self) -> ImageAndSpatialTransformerConfig:
        stn: SpatialTransformerNetwork = self.stn
        itn: Optional[ImageTransformerNetwork] = self.itn
        return ImageAndSpatialTransformerConfig(
            input=InputConfig(
                size=tuple(stn.in_size),
                channels=stn.in_channels // 2,
            ),
            itn=None if itn is None else itn.config,
            stn=stn.config,
        )

    @property
    def transform(self) -> GenericSpatialTransform:
        r"""Reference to spatial coordinates transformation."""
        return self.warp.transform

    def forward(
        self, source_img: Tensor, target_img: Tensor, apply: bool = True
    ) -> Dict[str, Tensor]:
        output: Dict[str, Tensor] = {}
        # Image transformation
        itn: Optional[Module] = self.itn
        source_soi = source_img
        target_soi = target_img
        if itn is not None:
            with torch.set_grad_enabled(itn.training):
                source_soi = itn(source_img)
                target_soi = itn(target_img)
            output["source_soi"] = source_soi
            output["target_soi"] = target_soi
        else:
            output["source_soi"] = source_img
            output["target_soi"] = target_img
        # Spatial transformation
        stn: Module = self.stn
        stn_input = torch.cat([source_soi, target_soi], dim=1)
        params: Dict[str, Tensor] = stn(stn_input)
        vfield_params: Optional[Tensor] = params.get("vfield")
        nonrigid_transform: Optional[ParametricTransform] = self.transform.get("nonrigid")
        if vfield_params is None:
            assert nonrigid_transform is None
        else:
            assert nonrigid_transform is not None
            vfield_shape = nonrigid_transform.data_shape[1:]
            vfield_params = U.grid_reshape(vfield_params, vfield_shape, align_corners=False)
            params["vfield"] = vfield_params
        self.transform.params = params
        if apply:
            output["warped_img"] = self.warp(source_img)
        return output


def create_istn(config: ImageAndSpatialTransformerConfig) -> ImageAndSpatialTransformerNetwork:
    itn = ImageTransformerNetwork(config.input.spatial_dims, config.input.channels, config.itn)
    itn_output_size = itn.output_size(config.input.size)
    stn = SpatialTransformerNetwork(2 * itn.out_channels, itn_output_size, config.stn)
    return ImageAndSpatialTransformerNetwork(itn=itn, stn=stn)



r"""Common types and functions that operate on tensors representing different types of data.

Besides defining common types and auxiliary functions extending the standard library,
this core library in particular defines functions which operate on objects of type ``torch.Tensor``.
This functional API defines a set of reusable state-less functions similar to ``torch.nn.functional``.
Object-oriented APIs use this functional API to realize their functionality. In particular, the
``forward()`` method of ``torch.nn.Module`` subclasses, such as for example data transformations
(cf. :mod:`.data.transforms`) and neural network components (cf. :mod:`.modules` and :mod:`.networks`)
are implemented using these functional building blocks.

The following import statement can be used to access the functional API:

.. code-block::

    import deepali.core.functional as U

"""

from importlib import metadata

from .enum import PaddingMode
from .enum import Sampling

from .grid import ALIGN_CORNERS
from .grid import Axes
from .cube import Cube
from .grid import Grid
from .grid import grid_points_transform
from .grid import grid_vectors_transform
from .grid import grid_transform_points
from .grid import grid_transform_vectors

from .pathlib import is_uri
from .pathlib import to_uri
from .storage import StorageObject

from .typing import Array
from .typing import Batch
from .typing import Dataclass
from .typing import Device
from .typing import DeviceStr
from .typing import DType
from .typing import DTypeStr
from .typing import Name
from .typing import PathStr
from .typing import PathUri
from .typing import Sample
from .typing import Scalar
from .typing import ScalarOrTuple
from .typing import ScalarOrTuple1d
from .typing import ScalarOrTuple2d
from .typing import ScalarOrTuple3d
from .typing import Size
from .typing import Shape
from .typing import TensorCollection

from .typing import is_bool_dtype
from .typing import is_float_dtype
from .typing import is_int_dtype
from .typing import is_uint_dtype
from .typing import is_namedtuple
from .typing import is_path_str


__version__ = metadata.version("hf-deepali")
r"""Version string of installed deepali core libraries."""

__all__ = (
    "ALIGN_CORNERS",
    "Array",
    "Axes",
    "Batch",
    "Cube",
    "Dataclass",
    "Device",
    "DeviceStr",
    "DType",
    "DTypeStr",
    "Grid",
    "Name",
    "PaddingMode",
    "PathStr",
    "PathUri",
    "Sample",
    "Sampling",
    "Scalar",
    "ScalarOrTuple",
    "ScalarOrTuple1d",
    "ScalarOrTuple2d",
    "ScalarOrTuple3d",
    "Size",
    "Shape",
    "StorageObject",
    "TensorCollection",
    "grid_points_transform",
    "grid_vectors_transform",
    "grid_transform_points",
    "grid_transform_vectors",
    "is_bool_dtype",
    "is_float_dtype",
    "is_int_dtype",
    "is_uint_dtype",
    "is_namedtuple",
    "is_path_str",
    "is_uri",
    "to_uri",
)



import pytest
import torch
from torch import Tensor

from deepali.core.kernels import gaussian1d
from deepali.core.random import _multinomial


@pytest.fixture
def generator() -> torch.Generator:
    return torch.Generator("cpu").manual_seed(123456789)


def test_multinomial_with_replacement(generator: torch.Generator) -> None:
    r"""Test weighted sampling with replacement using inverse transform sampling."""

    input = torch.arange(1000)

    # Input checks
    with pytest.raises(ValueError):
        _multinomial(torch.tensor(0), 1, replacement=True)
    with pytest.raises(ValueError):
        _multinomial(torch.ones((1, 2, 3)), 1, replacement=True)

    index = _multinomial(torch.ones(10), 11, replacement=True)  # no exception
    assert isinstance(index, Tensor)
    assert index.dtype == torch.int64
    assert index.shape == (11,)

    # Output type and shape
    index = _multinomial(input, 10, replacement=True)
    assert isinstance(index, Tensor)
    assert index.dtype == torch.int64
    assert index.shape == (10,)

    index = _multinomial(input.unsqueeze(0), 10, replacement=True)
    assert isinstance(index, Tensor)
    assert index.dtype == torch.int64
    assert index.shape == (1, 10)

    index = _multinomial(input.unsqueeze(0).repeat(3, 1), 10, replacement=True)
    assert isinstance(index, Tensor)
    assert index.dtype == torch.int64
    assert index.shape == (3, 10)

    # Only samples with non-zero weight
    subset = input.clone()
    subset[subset.lt(100)] = 0
    subset[subset.gt(200)] = 0

    for _ in range(10):
        index = _multinomial(subset, 10, replacement=True, generator=generator)
        assert isinstance(index, Tensor)
        assert index.dtype == torch.int64
        assert index.shape == (10,)
        assert subset[index].ge(100).le(200).all()

    # Sample from a spatial Gaussian distribution
    num_samples = 1_000_000
    input = gaussian1d(5)
    index = _multinomial(input, num_samples, replacement=True, generator=generator)
    assert isinstance(index, Tensor)
    assert index.dtype == torch.int64
    assert index.shape == (num_samples,)
    freq = index.bincount().div(num_samples)
    assert freq.allclose(input, rtol=0.1)


def test_multinomial_without_replacement(generator: torch.Generator) -> None:
    r"""Test weighted sampling without replacement using Gumbel-max trick."""

    input = torch.arange(1000)

    # Input checks
    with pytest.raises(ValueError):
        _multinomial(torch.tensor(0), 1, replacement=False)
    with pytest.raises(ValueError):
        _multinomial(torch.ones((1, 2, 3)), 1, replacement=False)

    index = _multinomial(torch.ones(10), 10, replacement=False)
    assert isinstance(index, Tensor)
    assert index.dtype == torch.int64
    assert index.shape == (10,)
    assert index.sort().values.eq(torch.arange(10)).all()

    with pytest.raises(ValueError):
        _multinomial(torch.ones(10), 11, replacement=False)

    # Output type and shape
    index = _multinomial(input, 10, replacement=False)
    assert isinstance(index, Tensor)
    assert index.dtype == torch.int64
    assert index.shape == (10,)

    index = _multinomial(input.unsqueeze(0), 10, replacement=False)
    assert isinstance(index, Tensor)
    assert index.dtype == torch.int64
    assert index.shape == (1, 10)

    index = _multinomial(input.unsqueeze(0).repeat(3, 1), 10, replacement=False)
    assert isinstance(index, Tensor)
    assert index.dtype == torch.int64
    assert index.shape == (3, 10)

    # No duplicates
    subset = input.clone()
    subset[subset.lt(100)] = 0
    subset[subset.gt(200)] = 0
    subset[subset.gt(0)] = 1

    num_samples = 10
    num_repeat = 100

    for _ in range(num_repeat):
        index = _multinomial(subset, num_samples, replacement=False, generator=generator)
        assert isinstance(index, Tensor)
        assert index.dtype == torch.int64
        assert index.shape == (num_samples,)
        assert index.unique().shape == (num_samples,)
        assert subset[index].eq(1).all()
